#/********************************************************************
# Cayenue/cayenue/panels/settings/settingspanel.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QGridLayout, QWidget, QTabWidget

from . import DiscoverOptions, GeneralOptions, StorageOptions, \
    AlarmOptions, ProxyOptions

class SettingsPanel(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw

        self.tab = QTabWidget()

        self.general = GeneralOptions(mw)
        self.discover = DiscoverOptions(mw)
        self.storage = StorageOptions(mw)
        self.proxy = ProxyOptions(mw)
        self.alarm = AlarmOptions(mw)

        self.tab.addTab(self.general, "General")
        self.tab.addTab(self.discover, "Discover")
        self.tab.addTab(self.storage, "Storage")
        self.tab.addTab(self.proxy, "Proxy")
        self.tab.addTab(self.alarm, "Alarm")

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.tab,   0, 0, 1, 1)

    def onMediaStarted(self, uri):
        if self.mw.pm.countPlayers():
            self.general.btnCloseAll.setText("Close All")

    def onMediaStopped(self, uri):
        if not self.mw.pm.countPlayers():
            self.general.btnCloseAll.setText("Start All")



#/********************************************************************
# Cayenue/cayenue/panels/settings/storage.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QMessageBox, QSpinBox, \
    QGridLayout, QWidget, QCheckBox, QLabel, QMessageBox, QGroupBox
from PyQt6.QtCore import QStandardPaths
from loguru import logger
from pathlib import Path
import shutil
from cayenue.components import DirectorySelector
from PyQt6.QtCore import pyqtSignal, QObject
import threading

class StorageSignals(QObject):
    updateDiskUsage = pyqtSignal()
    showWaitDialog = pyqtSignal(str)

class StorageOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.max_size = 100

        self.archiveKey = "settings/archive"
        self.pictureKey = "settings/picture"
        self.diskLimitKey = "settings/diskLimit"
        self.maxFileDurationKey = "settings/maxFileDuration"
        self.mangageDiskUsagekey = "settings/manageDiskUsage"
        self.writeBufferSizeKey = "setting/writeBufferSize"

        self.signals = StorageSignals()
        self.signals.updateDiskUsage.connect(self.updateDiskUsage)
        self.signals.showWaitDialog.connect(self.mw.dlgWait.show)

        video_dirs = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.MoviesLocation)
        self.dirArchive = DirectorySelector(mw, self.archiveKey, "Archive Dir", video_dirs[0])
        self.dirArchive.signals.dirChanged.connect(self.dirArchiveChanged)

        picture_dirs = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.PicturesLocation)
        self.dirPictures = DirectorySelector(mw, self.pictureKey, "Picture Dir", picture_dirs[0])
        self.dirPictures.signals.dirChanged.connect(self.dirPicturesChanged)

        self.grpDiskUsage = QGroupBox()

        self.chkManageDiskUsage = QCheckBox()
        self.chkManageDiskUsage.setChecked(bool(int(self.mw.settings.value(self.mangageDiskUsagekey, 0))))
        self.chkManageDiskUsage.clicked.connect(self.chkManageDiskUsageChanged)

        self.spnDiskLimit = QSpinBox()
        self.spnDiskLimit.valueChanged.connect(self.spnDiskLimitChanged)

        self.spnMaxFileDuration = QSpinBox()
        self.spnMaxFileDuration.setMaximum(60)
        self.spnMaxFileDuration.setValue(int(self.mw.settings.value(self.maxFileDurationKey, 15)))
        self.spnMaxFileDuration.valueChanged.connect(self.spnMaxFileDurationChanged)
        lblMaxFileDuration = QLabel("Max File Duration (minutes)")

        self.spnWriteBufferSize = QSpinBox()
        self.spnWriteBufferSize.setValue(int(self.mw.settings.value(self.writeBufferSizeKey, 10)))
        self.spnWriteBufferSize.valueChanged.connect(self.spnWriteBufferSizeChanged)
        lblWriteBuffferSize = QLabel("File Write Buffer Size (GB)")

        self.updateDiskUsage()
        #disk_limit = min(int(self.mw.settings.value(self.diskLimitKey, 100)), self.max_size)
        disk_limit = int(self.mw.settings.value(self.diskLimitKey, 100))
        self.spnDiskLimit.setValue(disk_limit)

        lytDiskUsage = QGridLayout(self.grpDiskUsage)
        lytDiskUsage.addWidget(self.chkManageDiskUsage, 0, 0, 1, 1)
        lytDiskUsage.addWidget(self.spnDiskLimit,       0, 2, 1, 1)
        lytDiskUsage.addWidget(QLabel("GB"),            0, 3, 1, 1)
        lytDiskUsage.addWidget(self.dirArchive,         1, 0, 1, 4)
        lytDiskUsage.addWidget(self.dirPictures,        2, 0, 1, 4)
        lytDiskUsage.addWidget(lblMaxFileDuration,      3, 0, 1, 2)
        lytDiskUsage.addWidget(self.spnMaxFileDuration, 3, 2, 1, 1)
        lytDiskUsage.addWidget(lblWriteBuffferSize,     4, 0, 1, 2)
        lytDiskUsage.addWidget(self.spnWriteBufferSize, 4, 2, 1, 1)
        lytDiskUsage.setColumnStretch(2, 10)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.grpDiskUsage, 0, 0, 1, 1)
        lytMain.addWidget(QLabel(),          1, 0, 1, 1)
        lytMain.addWidget(QLabel(),          2, 0, 1, 1)
        lytMain.setRowStretch(2, 10)


    def spnDiskLimitChanged(self, value):
        self.mw.settings.setValue(self.diskLimitKey, value)

    def spnMaxFileDurationChanged(self, value):
        self.mw.settings.setValue(self.maxFileDurationKey, value)
        self.mw.STD_FILE_DURATION = value * 60

    def chkManageDiskUsageChanged(self):
        if self.chkManageDiskUsage.isChecked():
            ret = QMessageBox.warning(self, "** WARNING **",
                                        "You are giving full control of both the video and picture archive directories to this program.  "
                                        "Any files contained within those directories or their sub-directories are subject to deletion.  "
                                        "You should only enable this feature if you are sure that this is ok.\n\n"
                                        "Are you sure you want to continue?",
                                        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
            if ret == QMessageBox.StandardButton.Cancel:
                self.chkManageDiskUsage.setChecked(False)
            else:
                self.signals.showWaitDialog.emit("Please wait for disk check to complete")
                thread = threading.Thread(target=self.calculateDiskUsage)
                thread.start()
                #self.updateDiskUsage()
        else:
            self.updateDiskUsage()
        self.mw.settings.setValue(self.mangageDiskUsagekey, int(self.chkManageDiskUsage.isChecked()))

    def dirArchiveChanged(self, path):
        logger.debug(f'Video archive directory changed to {path}')
        self.dirArchive.txtDirectory.setText(path)
        self.mw.settings.setValue(self.archiveKey, path)
        self.updateDiskUsage()
        #self.mw.filePanel.dirArchive.txtDirectory.setText(path)
        #self.mw.filePanel.dirChanged(path)

    def dirPicturesChanged(self, path):
        logger.debug(f'Picture directory changed to {path}')
        self.dirPictures.txtDirectory.setText(path)
        self.mw.settings.setValue(self.pictureKey, path)
        #self.mw.filePanel.control.dlgPicture.dirPictures.txtDirectory.setText(path)
        #self.mw.filePanel.control.dlgPicture.dirChanged(path)

    def spnWriteBufferSizeChanged(self, value):
        self.mw.settings.setValue(self.writeBufferSizeKey, value)
        self.updateDiskUsage()

    def calculateDiskUsage(self):
        if not self.mw.settings_profile == "gui":
            return
        dir = self.dirArchive.txtDirectory.text()
        size = sum(f.stat().st_size for f in Path(dir).rglob('*') if f.is_file())
        _, _, free = shutil.disk_usage(dir)
        buffer_size = self.spnWriteBufferSize.value() * 1_000_000_000
        self.max_size = int((free + size - buffer_size)/1_000_000_000)
        self.lblManage = f'Auto Manage (max {self.max_size} GB)'
        self.chkManageDiskUsage.setText(self.lblManage)
        self.spnDiskLimit.setMaximum(self.max_size)
        if self.spnDiskLimit.value() > self.max_size:
            self.spnDiskLimit.setValue(self.max_size)
        str_size_in_GB = "{:.2f}".format(size / 1_000_000_000)
        self.grpDiskUsage.setTitle(f'Disk Usage (currently {str_size_in_GB} GB)')
        self.mw.signals.hideWaitDialog.emit()

    def updateDiskUsage(self):
        if self.chkManageDiskUsage.isChecked():
            self.calculateDiskUsage()
        else:
            self.spnDiskLimit.setMaximum(self.max_size)
            self.grpDiskUsage.setTitle("Disk Usage Auto Manage")
            self.chkManageDiskUsage.setText("Auto Manage")


#/********************************************************************
# Cayenue/cayenue/panels/settings/proxy.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QMessageBox, QLineEdit, QPushButton, \
    QGridLayout, QWidget, QLabel, QMessageBox, QRadioButton, \
    QGroupBox, QCheckBox, QComboBox, QSpinBox
from PyQt6.QtCore import Qt
from cayenue.components import DirectorySelector
from cayenue.enums import ProxyType
from loguru import logger
import libonvif as onvif
import ipaddress
import os
import sys
import html

class ProxyOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw

        self.proxyTypeKey = "settings/proxyType"
        self.proxyRemoteKey = "settings/proxyRemote"
        self.autoDownloadKey = "settings/autoDownloadMTX"
        self.listenKey = "settings/alarmsFromServer"
        self.broadcastInterfaceKey = "settings/proxyBroadcastInterface"
        self.broadcastEnableKey = "setings/enableBroadcasting"
        self.broadcastAllInterfacesKey = "settings/broadcastAllInterfaces"
        self.timeoutKey = "settings/clientSocketTimeout"
        self.httpKey = "settings/enableHttpServer"

        self.if_addrs = []
        session = onvif.Session()
        session.getActiveInterfaces()
        i = 0
        while len(session.active_interface(i)) > 0 and i < 16:
            if session.active_interface(i) != "172.17.0.1":
                self.if_addrs.append(session.active_interface(i))
            i += 1

        self.grpProxyType = QGroupBox("Select Proxy Type")
        self.radStandAlone = QRadioButton("Stand Alone", self.grpProxyType)
        self.radServer = QRadioButton("Server", self.grpProxyType)
        self.radClient = QRadioButton("Client", self.grpProxyType)

        self.lblServer = QLabel("\n\n")
        self.lblConnect = QLabel("Connect url for clients")
        self.lblConnect.setEnabled(False)

        self.chkAutoDownload = QCheckBox("Auto Download Media MTX")
        self.chkAutoDownload.setEnabled(False)
        self.chkAutoDownload.setChecked(int(mw.settings.value(self.autoDownloadKey, 1)))
        self.chkAutoDownload.stateChanged.connect(self.chkAutoDownloadChecked)

        self.txtDirectoryMTX = DirectorySelector(mw, "MTXDir", "Dir", self.getProxyServerDir())
        self.txtDirectoryMTX.setEnabled(not self.chkAutoDownload.isChecked())

        self.cmbProxyLog = QComboBox()
        self.cmbProxyLog.addItems(["error", "warn", "info", "debug"])

        self.grpMediaMTX = QGroupBox("Media MTX Settings")
        lytMediaMTX = QGridLayout(self.grpMediaMTX)
        lytMediaMTX.addWidget(self.chkAutoDownload,       0, 0, 1, 2)
        lytMediaMTX.addWidget(self.txtDirectoryMTX,       1, 0, 1, 2)
        lytMediaMTX.addWidget(QLabel("Log Level"),        2, 0, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytMediaMTX.addWidget(self.cmbProxyLog,           2, 1, 1, 1)

        self.cmbInterfaces = QComboBox()
        self.lblInterfaces = QLabel("Network")

        # Windows can only broadcast on the network interface with highest priority
        # this needs further research
        if sys.platform == "win32":
            self.cmbInterfaces.addItem(session.primary_network_interface())
        else:
            for if_addr in self.if_addrs:
                self.cmbInterfaces.addItem(if_addr)
        
        if len(self.if_addrs):
            self.cmbInterfaces.setCurrentText(self.mw.settings.value(self.broadcastInterfaceKey, self.if_addrs[0]))
        else:
            logger.error("No active network interface was found")
        self.cmbInterfaces.currentTextChanged.connect(self.cmbInterfacesChanged)
        
        self.grpAlarmBroadcast = QGroupBox("Alarm Broadcasting")
        self.grpAlarmBroadcast.setCheckable(True)
        self.grpAlarmBroadcast.setChecked(int(self.mw.settings.value(self.broadcastEnableKey, 1)))
        self.grpAlarmBroadcast.clicked.connect(self.grpAlarmBroadcastClicked)
        lytAlarmBroadcast = QGridLayout(self.grpAlarmBroadcast)
        lytAlarmBroadcast.addWidget(self.lblInterfaces,              1, 0, 1, 1)
        lytAlarmBroadcast.addWidget(self.cmbInterfaces,              1, 1, 1, 1)

        self.proxyRemote = self.mw.settings.value(self.proxyRemoteKey)

        self.txtRemote = QLineEdit()
        self.txtRemote.setText(self.proxyRemote)
        self.txtRemote.textEdited.connect(self.txtRemoteEdited)
        self.txtRemote.setEnabled(False)
        self.lblRemote = QLabel("Connect url from server")
        self.lblRemote.setEnabled(False)

        self.btnUpdate = QPushButton("Update")
        self.btnUpdate.clicked.connect(self.btnUpdateClicked)
        self.btnUpdate.setEnabled(False)

        pnlRemote = QWidget()
        lytRemote = QGridLayout(pnlRemote)
        lytRemote.addWidget(self.btnUpdate, 0, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytRemote.addWidget(QLabel("   "),  0, 1, 1, 1)
        lytRemote.addWidget(self.txtRemote, 0, 2, 1, 1)
        lytRemote.setContentsMargins(10, 0, 10, 0)

        self.chkListen = QCheckBox("Get alarm events from server")
        self.chkListen.setChecked(int(mw.settings.value(self.listenKey, 1)))
        self.chkListen.stateChanged.connect(self.chkListenChecked)
        self.chkListen.setEnabled(False)

        self.spnTimeout = QSpinBox()
        self.spnTimeout.setValue(int(self.mw.settings.value(self.timeoutKey, 5)))
        self.spnTimeout.valueChanged.connect(self.spnTimeoutChanged)

        self.lblTimeout = QLabel("Client socket timeout (seconds)")
        pnlTimeout = QWidget()
        lytTimeout = QGridLayout(pnlTimeout)
        lytTimeout.addWidget(self.lblTimeout,  0, 0, 1, 1)
        lytTimeout.addWidget(self.spnTimeout,  0, 1, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytTimeout.setContentsMargins(0, 0, 0, 0)

        self.chkEnableHttpServer = QCheckBox("Enable HTTP server (port 8800)")
        self.chkEnableHttpServer.setChecked(int(mw.settings.value(self.httpKey, 0)))
        self.chkEnableHttpServer.stateChanged.connect(self.chkEnableHttpServerChecked)

        self.proxyType = None
        match self.mw.settings.value(self.proxyTypeKey, ProxyType.STAND_ALONE):
            case ProxyType.STAND_ALONE:
                self.radStandAlone.setChecked(True)
                self.radStandAloneToggled(True)
            case ProxyType.SERVER:
                self.radServer.setChecked(True)
                self.radServerToggled(True)
            case ProxyType.CLIENT:
                self.radClient.setChecked(True)
                self.radClientToggled(True)

        self.radStandAlone.toggled.connect(self.radStandAloneToggled)
        self.radServer.toggled.connect(self.radServerToggled)
        self.radClient.toggled.connect(self.radClientToggled)

        if self.mw.settings_profile != "gui":
            self.radServer.setEnabled(False)

        tag = "info"
        if self.proxyType == ProxyType.SERVER:
            try:
                filename = os.path.join(self.getProxyServerDir(), "mediamtx.yml") 
                with open(filename, 'r') as file:
                    content = file.read()
                    key = "logLevel:"
                    index = content.index(key)
                    eol = content.index("\n", index)
                    tag = content[index + len(key):eol].strip()
            except Exception as ex:
                logger.error(f"Media MTX configuration file read error: {ex}")
        self.cmbProxyLog.setCurrentText(tag)
        self.cmbProxyLog.currentTextChanged.connect(self.cmbProxyLogChanged)

        lytProxyType = QGridLayout(self.grpProxyType)
        lytProxyType.addWidget(self.radStandAlone,        0, 0, 1, 1)
        lytProxyType.addWidget(QLabel(),                  1, 0, 1, 2)
        lytProxyType.addWidget(self.radClient,            2, 0, 1, 1)
        lytProxyType.addWidget(self.lblRemote,            2, 1, 1, 1)
        lytProxyType.addWidget(pnlRemote,                 3, 0, 1, 2)
        lytProxyType.addWidget(self.chkListen,            4, 0, 1, 2)
        lytProxyType.addWidget(pnlTimeout,                5, 0, 1, 2)
        lytProxyType.addWidget(QLabel(),                  6, 0, 1, 2)
        lytProxyType.addWidget(self.radServer,            7, 0, 1, 1)
        lytProxyType.addWidget(self.lblConnect,           7, 1, 1, 1)
        lytProxyType.addWidget(self.lblServer,            9, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)
        lytProxyType.addWidget(self.chkEnableHttpServer, 11, 0, 1, 2)
        lytProxyType.addWidget(self.grpMediaMTX,         12, 0, 1, 2)
        lytProxyType.addWidget(self.grpAlarmBroadcast,   13, 0, 1, 2)
        lytProxyType.addWidget(QLabel(),                 15, 0, 1, 2)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.grpProxyType,  0, 0, 1, 2)

        #lytMain.addWidget(QLabel(),           2, 0, 1, 2)
        #lytMain.setRowStretch(2, 10)

    def getProxyServerDir(self):
        dir = None
        autoDownload = self.chkAutoDownload.isChecked()
        if autoDownload:
            dir = os.path.join(self.mw.getCacheLocation(), "proxy")
        else:
            dir = self.txtDirectoryMTX.text()
        return dir

    def setProxyType(self, type):
        if type == self.proxyType:
            return
        if self.mw.pm.countPlayers():
            ret = QMessageBox.question(self.mw, "Closing Streams", "Continuing with this operation will close all open streams and may require the application to restart.\n\nAre you sure you want to do this?")
            if ret == QMessageBox.StandardButton.No:
                match (self.proxyType):
                    case ProxyType.CLIENT:
                        self.radClient.setChecked(True)
                    case ProxyType.SERVER:
                        self.radServer.setChecked(True)
                    case ProxyType.STAND_ALONE:
                        self.radStandAlone.setChecked(True)
                return
            self.mw.closeAllStreams()

        self.proxyType = type
        self.mw.settings.setValue(self.proxyTypeKey, type)
        self.mw.stopProxyServer()
        self.mw.stopOnvifServer()
        self.mw.stopHttpServer()

        self.grpAlarmBroadcast.setEnabled(type == ProxyType.SERVER)
        self.grpMediaMTX.setEnabled(type == ProxyType.SERVER)
        self.chkEnableHttpServer.setEnabled(type == ProxyType.SERVER)

        if type == ProxyType.SERVER and self.mw.settings_profile == "gui":
            self.setServersLabel()
            self.mw.manageBroadcaster(self.getInterfaces())
            self.mw.startProxyServer(self.chkAutoDownload.isChecked(), self.getProxyServerDir())
            if self.chkEnableHttpServer.isChecked():
                self.mw.startHttpServer()
            self.mw.startOnvifServer("")

        if not hasattr(self.mw, "cameraPanel"):
            return

        if lstCamera := self.mw.cameraPanel.lstCamera:
            cameras = [lstCamera.item(x) for x in range(lstCamera.count())]
            for camera in cameras:
                self.mw.addCameraProxy(camera)
                if type == ProxyType.STAND_ALONE:
                    camera.onvif_data.nullifyGetProxyURI()
                else:
                    camera.onvif_data.getProxyURI = self.mw.getProxyURI
                for profile in camera.profiles:
                    if type == ProxyType.STAND_ALONE:
                        profile.nullifyGetProxyURI()
                    else:
                        profile.getProxyURI = self.mw.getProxyURI

        if type == ProxyType.CLIENT:
            self.manageAnalyzers(self.chkListen.isChecked())
            if len(self.mw.alarm_ordinals) != self.mw.cameraPanel.lstCamera.count():
                if not self.mw.cameraPanel.allCamerasFilled():
                    QMessageBox.warning(self.mw, "Application Restart Required", "It is necessary to re-start Onvif GUI in order to enable this change", QMessageBox.StandardButton.Close)
                    self.mw.close()
                    return
                self.mw.closeAllStreams()
                self.mw.cameraPanel.lstCamera.clear()
                self.mw.cameraPanel.btnDiscoverClicked()
        else:
            self.manageAnalyzers(False)

    def radClientToggled(self, checked):
        self.lblRemote.setEnabled(checked)
        self.txtRemote.setEnabled(checked)
        self.chkListen.setEnabled(checked)
        if checked:
           self.setProxyType(ProxyType.CLIENT)

    def radStandAloneToggled(self, checked):
        if checked:
            self.setProxyType(ProxyType.STAND_ALONE)

    def radServerToggled(self, checked):
        self.lblConnect.setEnabled(checked)
        self.chkAutoDownload.setEnabled(checked)
        if checked:
            self.setProxyType(ProxyType.SERVER)
            try:
                self.mw.cameraPanel.btnDiscoverClicked()
            except Exception as ex:
                pass
        else:
            self.lblServer.setText("\n\n")

    def setServersLabel(self):
        lbl = ""
        for if_addr in self.if_addrs:
            lbl += f'rtsp://{if_addr}:8554/\n'
        self.lblServer.setText(lbl)

    def txtRemoteEdited(self, arg):
        self.mw.settings.setValue(self.proxyRemoteKey, arg)
        comps = arg[:len(arg)-1][7:].split(":")
        ip_addr = comps[0]

        try:
            ip = ipaddress.ip_address(ip_addr)
            if self.txtRemote.text() != self.proxyRemote:
                self.btnUpdate.setEnabled(True)
        except ValueError:
            self.btnUpdate.setEnabled(False)

    def btnUpdateClicked(self):
        try:
            if self.mw.pm.countPlayers():
                QMessageBox.information(self.mw, "Closing Streams", "All current streams will be closed")
                self.mw.closeAllStreams()

            arg = self.txtRemote.text()
            comps = arg[:len(arg)-1][7:].split(":")
            ip_addr = comps[0]
            port = 8550
            if not self.mw.client:
                self.mw.initializeClient(ip_addr)
            else:
                self.mw.client.setEndpoint(ip_addr, port)
            self.proxyRemote = self.txtRemote.text()

            if lstCamera := self.mw.cameraPanel.lstCamera:
                cameras = [lstCamera.item(x) for x in range(lstCamera.count())]
                for camera in cameras:
                    self.mw.addCameraProxy(camera)

            self.mw.startListener(self.getInterfaces())

            self.btnUpdate.setEnabled(False)
        except Exception as ex:
            QMessageBox.critical(self.mw, "Proxy Client Error", str(ex))

    def chkAutoDownloadChecked(self, state):
        self.mw.settings.setValue(self.autoDownloadKey, state)
        self.txtDirectoryMTX.setEnabled(not self.chkAutoDownload.isChecked())

    def chkListenChecked(self, state):
        self.mw.settings.setValue(self.listenKey, state)
        self.manageAnalyzers(state)

    def chkEnableHttpServerChecked(self, state):
        self.mw.settings.setValue(self.httpKey, state)
        if state:
            self.createMtxIndex()
            self.mw.startHttpServer()
        else:
            self.mw.stopHttpServer()

    def spnTimeoutChanged(self, value):
        print("spnTimeoutChanged", value)
        self.mw.settings.setValue(self.timeoutKey, value)
        self.mw.client.timeout = value

    def grpAlarmBroadcastClicked(self):
        self.mw.settings.setValue(self.broadcastEnableKey, int(self.grpAlarmBroadcast.isChecked()))
        self.mw.manageBroadcaster(self.getInterfaces())

    def cmbInterfacesChanged(self, interface):
        self.mw.settings.setValue(self.broadcastInterfaceKey, interface)
        self.mw.manageBroadcaster(self.getInterfaces())

    def generateAlarmsLocally(self):
        if self.proxyType == ProxyType.CLIENT and self.chkListen.isChecked():
            return False
        else:
            return True
        
    def getTabIndexByName(self, name):
        index = -1
        for i in range(self.mw.tab.count()):
            if self.mw.tab.tabText(i) == name:
                return i
        return index
        
    def manageAnalyzers(self, state):
        if state:
            self.mw.tab.removeTab(self.getTabIndexByName("Video"))
            self.mw.tab.removeTab(self.getTabIndexByName("Audio"))
            self.mw.videoConfigure = None
            self.mw.audioConfigure = None
            self.mw.startListener(self.getInterfaces())
        else:
            self.mw.stopListener()
            self.mw.tab.addTab(self.mw.videoPanel, "Video")
            self.mw.tab.addTab(self.mw.audioPanel, "Audio")
            if not self.mw.videoConfigure:
                videoWorkerName = self.mw.videoPanel.cmbWorker.currentText()
                if len(videoWorkerName):
                    self.mw.loadVideoConfigure(videoWorkerName)
            if not self.mw.audioConfigure:
                audioWorkerName = self.mw.audioPanel.cmbWorker.currentText()
                if len(audioWorkerName):
                    self.mw.loadAudioConfigure(audioWorkerName)

    def getInterfaces(self):
        if_addrs = []
        match self.proxyType:
            case ProxyType.SERVER:
                if self.grpAlarmBroadcast.isChecked():
                    if_addrs.append(self.cmbInterfaces.currentText())
            case ProxyType.CLIENT:
                if_addrs = self.if_addrs
        return if_addrs
    
    def cmbProxyLogChanged(self, arg):
        try:
            filename = os.path.join(self.getProxyServerDir(), "mediamtx.yml") 
            with open(filename, 'r') as file:
                content = file.read()

            key = "logLevel:"
            index = content.index(key)
            eol = content.index("\n", index)
            tag = content[index:eol].strip()
            content = content.replace(tag, f'logLevel: {arg}')

            with open(filename, 'w') as file:
                file.write(content) 

        except Exception as ex:
            logger.error(f"Media MTX configuration file read error: {ex}")
        
    def setMediaMTXProxies(self):
        try:
            dir = self.getProxyServerDir()

            config_filename = f'{dir}/mediamtx.yml'
            if not os.path.isfile(config_filename):
                self.mw.signals.error.emit(f'Invalid path for MediaMTX configuration file: {config_filename}')
                return

            uri_map = {}
            for proxy_key in self.mw.proxies:
                username = None
                password = None
                if camera := self.mw.cameraPanel.getCamera(self.mw.proxies[proxy_key]):
                    if profile := camera.getDisplayProfile():
                        username = profile.username()
                        password = profile.password()
                mtx_key = self.mw.proxies[proxy_key][len("rtsp://"):]
                mtx_key = mtx_key[mtx_key.find('/')+1:]

                uri = None
                if username and password:
                    uri = f'rtsp://{username}:{password}@{proxy_key[len("rtsp://"):]}'
                else:
                    uri = proxy_key

                uri_map[mtx_key] = uri                

            mtx_uris = {}
            relevant = False
            key = None
            value = None
            with open(config_filename, 'r') as config:
                for readline in config:
                    line = readline.strip()

                    if relevant:
                        if not line.startswith("#"):
                            if not key:
                                key = line[:-1]
                            else:
                                value = line[len("source: "):]
                                mtx_uris[key] = value
                                key = None
                                value = None

                    if line.startswith("paths:"):
                        relevant = True

                    if line.startswith("all_others"):
                        relevant = False

            if mtx_uris == uri_map:
                return

            relevant = False
            lines = []
            with open(config_filename, 'r') as config:
                for readline in config:
                    line = readline.strip()

                    if not relevant or line.startswith("#"):
                        lines.append(readline)

                    if line.startswith("paths:"):
                        relevant = True

                        for key in uri_map:
                            lines.append(f'  {key}:\n')
                            lines.append(f'    source: {uri_map[key]}\n')

                    if line.startswith("all_others"):
                        lines.append(readline)
                        relevant = False

            
            with open(config_filename, 'w') as config:
                for line in lines:
                    config.write(line)

            self.createMtxIndex()

        except Exception as ex:
            logger.error(f'Error setting Media MTX proxies: {ex}')
            self.mw.signals.error.emit(f'Error setting Media MTX proxies: {ex}')

    def createMtxIndex(self):
        try:
            links = {}
            lst = self.mw.cameraPanel.lstCamera
            cameras = [lst.item(x) for x in range(lst.count())]
            for camera in cameras:
                for profile in camera.profiles:
                    links[f"{camera.name()} ({profile.profile()})"] = f"{camera.serial_number()}/{profile.profile()}"

            with open(f'{self.getProxyServerDir()}/index.html', 'w') as file:
                file.write(self.generateIndexHtml(links))

        except Exception as ex:
            logger.error(f"error generating MTX index.html: {ex}")

    def generateIndexHtml(self, cameras, webrtc_port=8889):
        links_html = "\n".join(
            f'    <li><a class="cam" data-path="{html.escape(path)}">'
            f'{html.escape(name)}</a></li>'
            for name, path in cameras.items()
        )

        return f"""<!DOCTYPE html>
    <html>
    <head>
    <meta charset="utf-8">
    <title>Cameras</title>
    </head>
    <body>

    <h1>Camera List</h1>

    <ul>
    {links_html}
    </ul>

    <script>
    const webrtcPort = {webrtc_port};
    const loc = window.location;

    document.querySelectorAll(".cam").forEach(a => {{
        const path = a.dataset.path;
        a.href = `${{loc.protocol}}//${{loc.hostname}}:${{webrtcPort}}/${{path}}`;
    }});
    </script>

    </body>
    </html>
    """



#/********************************************************************
# Cayenue/cayenue/panels/camera/camerapanel.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from time import sleep
from PyQt6.QtWidgets import QPushButton, QGridLayout, QWidget, QSlider, \
    QTabWidget, QMessageBox, QMenu, QFileDialog
from PyQt6.QtGui import QAction
from PyQt6.QtCore import Qt, pyqtSignal, QObject, QTimer, QSettings
from . import NetworkTab, ImageTab, VideoTab, PTZTab, SystemTab, LoginDialog, \
    Session, Camera, CameraList, Snapshot
from loguru import logger
import libonvif as onvif
from pathlib import Path
import os
import subprocess
from datetime import datetime
from cayenue.enums import ProxyType
import platform
import webbrowser
import threading
import sys
import time

class CameraPanelSignals(QObject):
    fill = pyqtSignal(onvif.Data)
    login = pyqtSignal(onvif.Data)
    collapseSplitter = pyqtSignal()
    guiSync = pyqtSignal()

class CameraPanel(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.dlgLogin = LoginDialog(self)
        self.fillers = []
        self.sync_lock = False
        self.snapshot = Snapshot(mw)

        self.autoTimeSyncer = None
        self.enableAutoTimeSync(self.mw.settingsPanel.general.chkAutoTimeSync.isChecked())
       
        self.lstCamera = CameraList(mw)
        self.lstCamera.currentItemChanged.connect(self.onCurrentItemChanged)
        self.lstCamera.itemDoubleClicked.connect(self.onItemDoubleClicked)
        self.lstCamera.itemClicked.connect(self.onItemClicked)
        self.lstCamera.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.lstCamera.customContextMenuRequested.connect(self.showContextMenu)

        self.sldVolume = QSlider(Qt.Orientation.Horizontal)
        self.sldVolume.setValue(80)
        self.sldVolume.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.sldVolume.valueChanged.connect(self.sldVolumeChanged)
        self.sldVolume.setEnabled(False)

        self.btnStop = QPushButton()
        self.btnStop.setMinimumWidth(40)
        self.btnStop.setMaximumHeight(20)
        self.btnStop.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnStop.clicked.connect(self.btnStopClicked)

        self.btnRecord = QPushButton()
        self.btnRecord.setMinimumWidth(40)
        self.btnRecord.setMaximumHeight(20)
        self.btnRecord.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnRecord.clicked.connect(self.btnRecordClicked)

        self.btnMute = QPushButton()
        self.btnMute.setMinimumWidth(40)
        self.btnMute.setMaximumHeight(20)
        self.btnMute.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnMute.clicked.connect(self.btnMuteClicked)

        self.btnDiscover = QPushButton()
        self.btnDiscover.setMinimumWidth(40)
        self.btnDiscover.setMaximumHeight(20)
        self.btnDiscover.setStyleSheet(self.getButtonStyle("discover"))
        self.btnDiscover.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnDiscover.clicked.connect(self.btnDiscoverClicked)

        self.btnApply = QPushButton()
        self.btnApply.setMinimumWidth(40)
        self.btnApply.setMaximumHeight(20)
        self.btnApply.setStyleSheet(self.getButtonStyle("apply"))
        self.btnApply.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnApply.clicked.connect(self.btnApplyClicked)
        self.btnApply.setEnabled(False)
        
        self.btnSnapshot = QPushButton()
        self.btnSnapshot.setMinimumWidth(40)
        self.btnSnapshot.setMaximumHeight(20)
        self.btnSnapshot.setStyleSheet(self.getButtonStyle("snapshot"))
        self.btnSnapshot.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnSnapshot.clicked.connect(self.btnSnapshotClicked)
        self.btnSnapshot.setEnabled(False)        

        self.btnHelp = QPushButton()
        self.btnHelp.setMinimumWidth(40)
        self.btnHelp.setMaximumHeight(20)
        self.btnHelp.setStyleSheet(self.getButtonStyle("help"))
        self.btnHelp.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnHelp.clicked.connect(self.btnHelpClicked)
        
        self.btnStopAll = QPushButton()
        self.btnStopAll.setMinimumWidth(40)
        self.btnStopAll.setMaximumHeight(20)
        self.btnStopAll.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnStopAll.clicked.connect(self.btnStopAllClicked)

        self.btnHistory = QPushButton()
        self.btnHistory.setMinimumWidth(40)
        self.btnHistory.setMaximumHeight(20)
        self.btnHistory.setStyleSheet(self.getButtonStyle("history"))
        self.btnHistory.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnHistory.clicked.connect(self.btnHistoryClicked)

        self.btnFullScreen = QPushButton()
        self.btnFullScreen.setMinimumWidth(40)
        self.btnFullScreen.setMaximumHeight(20)
        self.btnFullScreen.setStyleSheet(self.getButtonStyle("full_screen"))
        self.btnFullScreen.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnFullScreen.clicked.connect(self.btnFullScreenClicked)

        self.tabOnvif = QTabWidget()
        self.tabOnvif.setUsesScrollButtons(False)
        self.tabVideo = VideoTab(self)
        self.tabImage = ImageTab(self)
        self.tabNetwork = NetworkTab(self)
        self.ptzTab = PTZTab(self)
        self.tabSystem = SystemTab(self)
        self.tabOnvif.addTab(self.tabVideo, "Media")
        self.tabOnvif.addTab(self.tabImage, "Image")
        self.tabOnvif.addTab(self.tabNetwork, "Network")
        self.tabOnvif.addTab(self.ptzTab, "PTZ")
        self.tabOnvif.addTab(self.tabSystem, "System")

        self.signals = CameraPanelSignals()
        self.signals.fill.connect(self.tabVideo.fill)
        self.signals.fill.connect(self.tabImage.fill)
        self.signals.fill.connect(self.tabNetwork.fill)
        self.signals.fill.connect(self.ptzTab.fill)
        self.signals.fill.connect(self.tabSystem.fill)
        self.signals.fill.connect(self.syncGUI)
        self.signals.login.connect(self.onShowLogin)
        self.signals.collapseSplitter.connect(self.mw.collapseSplitter)
        self.signals.guiSync.connect(self.syncGUI)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.lstCamera,     0, 0, 1, 6)
        lytMain.addWidget(self.tabOnvif,      1, 0, 1, 6)
        lytMain.addWidget(self.btnStopAll,    2, 0, 1, 1)
        lytMain.addWidget(self.btnHistory,    2, 1, 1, 1)
        lytMain.addWidget(self.btnSnapshot,   2, 2, 1, 1)
        lytMain.addWidget(self.btnFullScreen, 2, 3, 1, 1)
        lytMain.addWidget(self.btnHelp,       2, 4, 1, 1)
        lytMain.addWidget(self.btnStop,       3, 0, 1, 1)
        lytMain.addWidget(self.btnRecord,     3, 1, 1, 1)
        lytMain.addWidget(self.btnDiscover,   3, 2, 1, 1)
        lytMain.addWidget(self.btnApply,      3, 3, 1, 1)
        lytMain.addWidget(self.btnMute,       3, 4, 1, 1)
        lytMain.addWidget(self.sldVolume,     3, 5, 1, 1)
        lytMain.setColumnStretch(5, 10)
        lytMain.setRowStretch(0, 10)

        self.menu = QMenu("Context Menu", self)
        self.remove = QAction("Delete", self)
        self.rename = QAction("Rename", self)
        self.info = QAction("Info", self)
        self.password = QAction("Password", self)
        self.start = QAction("Start", self)
        self.stop = QAction("Stop", self)
        self.remove.triggered.connect(self.onMenuRemove)
        self.rename.triggered.connect(self.onMenuRename)
        self.info.triggered.connect(self.onMenuInfo)
        self.password.triggered.connect(self.onMenuPassword)
        self.start.triggered.connect(self.onMenuStart)
        self.stop.triggered.connect(self.onMenuStop)
        self.menu.addAction(self.remove)
        self.menu.addAction(self.rename)
        self.menu.addAction(self.info)
        self.menu.addAction(self.password)
        self.menu.addAction(self.start)
        self.menu.addAction(self.stop)

        self.syncGUI()
        self.setTabsEnabled(False)
        self.sessions = []
        self.closing = False

    def showContextMenu(self, pos):
        index = self.lstCamera.indexAt(pos)
        if index.isValid():
            self.menu.exec(self.mapToGlobal(pos))

    def onMenuRemove(self):
        self.lstCamera.remove()
    
    def onMenuRename(self):
        self.lstCamera.rename()

    def onMenuInfo(self):
        self.lstCamera.info()

    def onMenuPassword(self):
        self.lstCamera.password()

    def onMenuStart(self):
        self.lstCamera.startCamera()

    def onMenuStop(self):
        self.lstCamera.stopCamera()

    def btnDiscoverClicked(self):
        if self.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
            if self.mw.client:
                self.mw.client.transmit(bytearray("GET CAMERAS\r\n", 'utf-8'))
            return
        
        if self.mw.settingsPanel.discover.radDiscover.isChecked():
            logger.debug("Using broadcast discovery")
            interfaces = []
            self.sessions.clear()

            if self.mw.settingsPanel.discover.chkScanAllNetworks.isChecked():
                for i in range(self.mw.settingsPanel.discover.cmbInterfaces.count()):
                    interfaces.append(self.mw.settingsPanel.discover.cmbInterfaces.itemText(i))
            else:
                interfaces.append(self.mw.settingsPanel.discover.cmbInterfaces.currentText())

            for interface in interfaces:
                session = Session(self, interface)
                session.start()
                self.sessions.append(session)
                self.btnDiscover.setEnabled(False)
                if len(interfaces) > 1:
                    sleep(1)
        else:
            logger.debug("Using cached camera addresses for discovery")
            self.fillers.clear()
            tmp = self.mw.settings.value(self.mw.settingsPanel.discover.cameraListKey)
            if tmp:
                numbers = tmp.strip().split("\n")
                for serial_number in numbers:
                    key = f'{serial_number}/XAddrs'
                    xaddrs = self.mw.settings.value(key)
                    alias = self.mw.settings.value(f'{serial_number}/Alias')
                    data = onvif.Data()
                    data.errorCallback = self.errorCallback
                    data.infoCallback = self.infoCallback
                    data.setSetting = self.mw.settings.setValue
                    data.getSetting = self.mw.settings.value
                    data.getData = self.getData
                    data.getCredential = self.getCredential
                    data.setXAddrs(xaddrs)
                    data.alias = alias
                    data.setCameraName(alias)
                    data.setDeviceService("POST /onvif/device_service HTTP/1.1\r\n")
                    self.fillers.append(data)
                    data.startManualFill()

    def errorCallback(self, msg):
        try:
            self.mw.signals.error.emit(msg)
        except Exception as ex:
            logger.error("camera panel error callback exception: {ex}")

    def infoCallback(self, msg):
        try:
            if msg.startswith("Set System Date and Time Error"):
                logger.error(msg)
            else:
                logger.debug(msg)
        except Exception as ex:
            logger.error("camera panel info callback exception: {ex}")

    def discoveryTimeout(self):
        self.setEnabled(True)
        self.btnDiscover.setEnabled(True)

    def discovered(self):
        finished = True
        for session in self.sessions:
            if session.active:
                finished = False
                break

        if finished:
            self.sessions.clear()
            self.btnDiscover.setEnabled(True)
            if self.mw.settingsPanel.proxy.proxyType == ProxyType.SERVER:
                self.mw.settingsPanel.proxy.setMediaMTXProxies()

    def getCredential(self, onvif_data):
        if not onvif_data: return
        
        if self.getCameraByXAddrs(onvif_data.xaddrs()) and not len(self.fillers):
            onvif_data.cancelled = True
            return onvif_data
                
        alternateUsername = self.mw.settings.value(f'{onvif_data.xaddrs()}/alternateUsername', None)
        alternatePassword = self.mw.settings.value(f'{onvif_data.xaddrs()}/alternatePassword', None)

        if (len(self.mw.settingsPanel.general.txtPassword.text()) or alternatePassword) and len(onvif_data.last_error()) == 0:
            if alternateUsername:
                onvif_data.setUsername(alternateUsername)
            else:
                onvif_data.setUsername(self.mw.settingsPanel.general.txtUsername.text())
            if alternatePassword:
                onvif_data.setPassword(alternatePassword)
            else:
                onvif_data.setPassword(self.mw.settingsPanel.general.txtPassword.text())
        else:
            if onvif_data.last_error().startswith("Network error, unable to connect"):
                logger.debug(f'Unable to connect with {onvif_data.xaddrs()}')
                onvif_data.cancelled = True
                self.mw.signals.error.emit(f'Unable to connect with {onvif_data.xaddrs()}')
            else:
                while self.dlgLogin.active:
                    sleep(0.01)

                self.dlgLogin.active = True
                self.signals.login.emit(onvif_data)
                while self.dlgLogin.active:
                    sleep(0.01)

        return onvif_data
    
    def onShowLogin(self, onvif_data):
        self.dlgLogin.exec(onvif_data)

    def getProxyData(self, onvif_data):
        if not onvif_data: return
        
        onvif_data.getProxyURI = self.mw.getProxyURI

        alias = self.mw.settings.value(f'{onvif_data.serial_number()}/Alias')
        if not alias:
            name = onvif_data.camera_name()
            if len(name):
                alias = name
                self.mw.settings.setValue(f'{onvif_data.serial_number()}/Alias', name)
            else:
                alias = onvif_data.host()
        onvif_data.alias = alias

        if not self.getCameraBySerialNumber(onvif_data.serial_number()):

            self.mw.alarm_ordinals[len(self.mw.alarm_ordinals)] = onvif_data.serial_number()

            add_camera = True
            if self.mw.settingsPanel.discover.radCached.isChecked() and self.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                if tmp := self.mw.settings.value(self.mw.settingsPanel.discover.cameraListKey):
                    numbers = tmp.strip().split("\n")
                    if onvif_data.serial_number() not in numbers:
                        add_camera = False

            if add_camera:
                camera = Camera(onvif_data, self.mw)
                camera.setIconIdle()
                camera.dimForeground()
                self.mw.addCameraProxy(camera)
                self.lstCamera.addItem(camera)
                self.lstCamera.sortItems()
                camera.setDisplayProfile(camera.getDisplayProfileSetting())
                logger.debug(f'Discovery completed for Camera: {onvif_data.alias}, Stream URI: {onvif_data.stream_uri()}, xaddrs: {onvif_data.xaddrs()}')

                self.filled(onvif_data)

    def getData(self, onvif_data):
        if not onvif_data: return
        
        if onvif_data.last_error().startswith("Error initializing camera data during manual fill:"):
            logger.debug(onvif_data.last_error())
            return

        onvif_data.filled = self.filled
        onvif_data.infoCallback = self.infoCallback
        onvif_data.errorCallback = self.errorCallback

        alias = self.mw.settings.value(f'{onvif_data.serial_number()}/Alias')
        if not alias:
            name = onvif_data.camera_name()
            if len(name):
                alias = name
                self.mw.settings.setValue(f'{onvif_data.serial_number()}/Alias', name)
            else:
                alias = onvif_data.host()
        onvif_data.alias = alias

        if existing := self.getCameraBySerialNumber(onvif_data.serial_number()):
            synchronizeTime = self.mw.settingsPanel.general.chkAutoTimeSync.isChecked()
            if not self.closing:
                existing.onvif_data.setXAddrs(onvif_data.xaddrs())
                for profile in existing.profiles:
                    profile.setXAddrs(onvif_data.xaddrs())
                existing.onvif_data.startFill(synchronizeTime)
        else:
            camera = Camera(onvif_data, self.mw)
            camera.setIconIdle()
            camera.dimForeground()
            self.mw.addCameraProxy(camera)
            
            self.lstCamera.addItem(camera)
            self.lstCamera.sortItems()
            camera.setDisplayProfile(camera.getDisplayProfileSetting())
            self.saveCameraList()
            logger.debug(f'Discovery completed for Camera: {onvif_data.alias}, Stream URI: {onvif_data.stream_uri()}, xaddrs: {onvif_data.xaddrs()}, {onvif_data.camera_name()}')

            synchronizeTime = self.mw.settingsPanel.general.chkAutoTimeSync.isChecked()
            if not self.closing:
                onvif_data.startFill(synchronizeTime)

    def filled(self, onvif_data):
        if not onvif_data: return
        
        if camera := self.getCamera(onvif_data.uri()):
            camera.restoreForeground()
            key = f'{camera.serial_number()}/XAddrs'
            self.mw.settings.setValue(key, camera.xaddrs())

            if camera.manual_fill:
                camera.assignData(onvif_data)
                self.mw.addCameraProxy(camera)
                camera.setDisplayProfile(camera.getDisplayProfileSetting())

            if self.lstCamera is not None:
                current_camera = self.getCurrentCamera()
                if current_camera:
                    if current_camera.xaddrs() == onvif_data.xaddrs():
                        self.signals.fill.emit(onvif_data)
                        self.setEnabled(True)
                        self.setTabsEnabled(True)

            camera.filled = True
            if self.allCamerasFilled():
                self.signals.guiSync.emit()
            if self.mw.settingsPanel.proxy.proxyType == ProxyType.SERVER and \
                    self.mw.settingsPanel.discover.radCached.isChecked() and \
                    self.allCamerasFilled():
                
                self.mw.settingsPanel.proxy.setMediaMTXProxies()

            # auto start after fill, recording needs onvif frame rate
            if self.mw.settingsPanel.discover.chkAutoStart.isChecked():
                if not camera.isRunning():
                    while not self.mw.isVisible():
                        sleep(0.1)
                    self.lstCamera.itemClicked.emit(camera)
                    self.lstCamera.itemDoubleClicked.emit(camera)
                    sleep(0.1)

        if len(onvif_data.last_error()):
            logger.debug(f'Error from {onvif_data.alias} : {onvif_data.last_error()}')

    def saveCameraList(self):
        serial_numbers = ""
        cameras = [self.mw.cameraPanel.lstCamera.item(x) for x in range(self.mw.cameraPanel.lstCamera.count())]
        for camera in cameras:
            serial_numbers += camera.serial_number() + "\n"
        self.mw.settings.setValue(self.mw.settingsPanel.discover.cameraListKey, serial_numbers)

    def onCurrentItemChanged(self, current, previous):
        if current:
            if self.mw.pm.getPlayer(current.uri()):
                self.mw.glWidget.focused_uri = current.uri()
            else:
                self.mw.glWidget.focused_uri = None
            self.signals.fill.emit(current.onvif_data)
            self.syncGUI()

    def onItemClicked(self, camera):
        if self.mw.videoConfigure:
            self.mw.videoConfigure.setCamera(camera)

    def onItemDoubleClicked(self, camera):
        if not camera: return
        profiles = self.mw.pm.getStreamPairProfiles(camera.uri())
        players = self.mw.pm.getStreamPairPlayers(camera.uri())
        timers = self.mw.pm.getStreamPairTimers(camera.uri())

        activeTimer = False
        for timer in timers:
            if timer.isActive():
                activeTimer = True

        if activeTimer:
            for timer in timers:
                self.mw.signals.stopReconnect.emit(timer.uri)
            for player in players:
                player.requestShutdown()
            camera.setIconIdle()
        else:
            if len(players):
                if player := self.mw.cameraPanel.getCurrentPlayer():
                    if profile := camera.getRecordProfile():
                        self.mw.openFocusWindow()
                        count = 0
                        while not self.mw.focus_window.cameraPanel.getCamera(profile.uri()):
                            time.sleep(0.01)
                            count += 1
                            if count > 200:
                                logger.error("timeout error opening focus window")
                                break
                        if camera := self.mw.focus_window.cameraPanel.getCamera(profile.uri()):
                            self.mw.focus_window.cameraPanel.onItemDoubleClicked(camera)
            else:
                for i, profile in enumerate(profiles):
                    if i == 0:
                        profile.setHidden(False)
                        self.mw.playMedia(profile.uri())
                    else:
                        if camera.displayProfileIndex() != camera.recordProfileIndex():
                            profile.setHidden(True)
                            self.mw.playMedia(profile.uri())

        self.syncGUI()

    def setTabsEnabled(self, enabled):
        self.tabVideo.setEnabled(enabled)
        self.tabImage.setEnabled(enabled)
        self.tabNetwork.setEnabled(enabled)
        self.ptzTab.setEnabled(enabled)
        self.tabSystem.setEnabled(enabled)

    def btnApplyClicked(self):
        if camera := self.getCurrentCamera():
            self.btnApply.setEnabled(False)
            self.tabVideo.update(camera.onvif_data)
            self.tabImage.update(camera.onvif_data)
            self.tabNetwork.update(camera.onvif_data)

    def onEdit(self):
        if camera := self.getCurrentCamera():
            if self.tabVideo.edited(camera.onvif_data) or \
                    self.tabImage.edited(camera.onvif_data) or \
                    self.tabNetwork.edited(camera.onvif_data):
                self.btnApply.setEnabled(True)
            else:
                self.btnApply.setEnabled(False)

    def sldVolumeChanged(self, value):
        if player := self.getCurrentPlayer():
            player.setVolume(value)
        if camera := self.getCurrentCamera():
            camera.setVolume(value)

    def btnMuteClicked(self):
        player = self.getCurrentPlayer()
        if player:
            player.setMute(not player.isMuted())
            camera = self.getCurrentCamera()
            if camera:
                camera.setMute(player.isMuted())
        else:
            camera = self.getCurrentCamera()
            if camera:
                camera.setMute(not camera.mute)
        self.syncGUI()

    def btnRecordClicked(self):
        player = self.getCurrentPlayer()
        camera = self.getCurrentCamera()
        if camera.displayProfileIndex() != camera.recordProfileIndex():
            recordProfile = camera.getRecordProfile()
            if recordProfile:
                record_uri = recordProfile.uri()
                player = self.mw.pm.getPlayer(record_uri)

        if player:
            if player.isRecording():
                player.output_file_start_time = None
                player.toggleRecording("")
                self.mw.settingsPanel.storage.signals.updateDiskUsage.emit()
                if camera:
                    camera.manual_recording = False
            else:
                #d = self.mw.settingsPanel.storage.dirArchive.txtDirectory.text()
                #if self.mw.settingsPanel.storage.chkManageDiskUsage.isChecked():
                #    self.mw.diskManager.manageDirectory(d)
                #else:
                #    self.mw.settingsPanel.storage.signals.updateDiskUsage.emit()
                if filename := player.getOutputFilename():
                    player.toggleRecording(filename)
                    if camera:
                        camera.manual_recording = True

        self.syncGUI()

    def btnHistoryClicked(self):
        try:
            reader_settings = QSettings("Cayenue", "Reader")
            reader_settings.setValue("filePanel/hideCameraPanel", 1)
            main_file = Path(__file__).parent.parent.parent / "main.py"
            #subprocess.Popen(["python", str(main_file), "--profile", "reader"], env=os.environ.copy(), start_new_session=True, shell=True)
            if platform.system() == "Windows":
                subprocess.Popen([sys.executable, str(main_file), "--profile", "Reader"], env=os.environ.copy(), start_new_session=True, shell=True)
            else:
                subprocess.Popen([sys.executable, str(main_file), "--profile", "Reader"], env=os.environ.copy(), start_new_session=True)
        except Exception as ex:
            logger.error(f'Error starting file browser: {ex}')
            self.mw.signals.error.emit(f'Error starting file browser: {ex}')

    def btnStopClicked(self):
        if camera := self.getCurrentCamera():
            #profiles = self.mw.pm.getStreamPairProfiles(camera.uri())
            players = self.mw.pm.getStreamPairPlayers(camera.uri())
            timers = self.mw.pm.getStreamPairTimers(camera.uri())

            activeTimer = False
            for timer in timers:
                if timer.isActive():
                    activeTimer = True

            if activeTimer:
                for timer in timers:
                    self.mw.signals.stopReconnect.emit(timer.uri)
                for player in players:
                    player.requestShutdown()
                camera.setIconIdle()
                self.syncGUI()
                return

            if len(players):
                for player in players:
                    player.requestShutdown()
                camera.setIconIdle()
                self.syncGUI()
                return
            else:
                self.onItemDoubleClicked(camera)
                self.syncGUI()

    def btnHelpClicked(self):
        result = webbrowser.get().open("https://github.com/sr99622/libonvif#readme-ov-file")
        if not result:
            webbrowser.get().open("https://github.com/sr99622/libonvif")

    def btnSnapshotClicked(self):
        if player := self.getCurrentPlayer():
            root = self.mw.settingsPanel.storage.dirPictures.txtDirectory.text() + "/" + self.getCamera(player.uri).text()
            Path(root).mkdir(parents=True, exist_ok=True)
            filename = '{0:%Y%m%d%H%M%S.jpg}'.format(datetime.now())
            filename = str(root + "/" + filename)

            if self.mw.settingsPanel.general.chkSnapshotDlg.isChecked():
                if platform.system() == "Linux":
                    filename = QFileDialog.getSaveFileName(self, "Save File As", filename, options=QFileDialog.Option.DontUseNativeDialog)[0]
                else:
                    filename = QFileDialog.getSaveFileName(self, "Save File As", filename)[0]

            if len(Path(filename).stem):
                answer = QMessageBox.StandardButton.Yes
                if not filename.endswith(".jpg"):
                    filename += ".jpg"
                    if Path(filename).is_file():
                        answer = QMessageBox.question(self.mw, "File Exists", "You are about to overwrite an existing file, are you sure you wnat to do this?")
                if answer == QMessageBox.StandardButton.Yes:
                    if camera := self.getCamera(player.uri):
                        profile = camera.getRecordProfile()
                        if not profile:
                            profile = camera.getProfile(player.uri)

                        thread = threading.Thread(target=self.snapshot, args=(profile, filename, camera, player))
                        thread.start()

    def btnStopAllClicked(self):
        if len(self.mw.pm.players):
            self.mw.closeAllStreams()
        else:
            self.mw.startAllCameras()
        self.syncGUI()

    def btnFullScreenClicked(self):
        if self.mw.isFullScreen():
            self.mw.showNormal()
            self.btnFullScreen.setStyleSheet(self.getButtonStyle("full_screen"))
        else:
            self.mw.showFullScreen()
            self.btnFullScreen.setStyleSheet(self.getButtonStyle("normal"))
       
    def onMediaStarted(self, uri):
        if self.mw.tabVisible:
            if self.mw.glWidget.focused_uri is None:
                if camera := self.getCurrentCamera():
                    self.mw.glWidget.focused_uri = camera.uri()

        #self.tabVideo.btnSnapshot.setEnabled(True)
        
        self.syncGUI()

    def onMediaStopped(self, uri):
        if camera := self.getCamera(uri):
            camera.setIconIdle()
            if profile := camera.getProfile(camera.uri()):
                if profile.getAnalyzeVideo():
                    if self.mw.videoWorker:
                        self.mw.videoWorker(None, None)
                if profile.getAnalyzeAudio():
                    if self.mw.audioWorker:
                        self.mw.audioWorker(None, None)
        #self.tabVideo.btnSnapshot.setEnabled(False)
        self.syncGUI()

    def syncGUI(self):
        while (self.sync_lock):
            sleep(0.001)
        self.sync_lock = True
        
        if camera := self.getCurrentCamera():
            self.btnStop.setEnabled(True)
            if player := self.mw.pm.getPlayer(camera.uri()):
                self.btnStop.setStyleSheet(self.getButtonStyle("stop"))
                self.btnSnapshot.setEnabled(True)
                #self.tabVideo.btnSnapshot.setEnabled(True)

                if ps := player.systemTabSettings():
                    self.btnRecord.setEnabled(not (ps.record_enable and ps.record_always))

                if player.hasAudio() and not player.disable_audio:
                    self.btnMute.setEnabled(True)
                    self.sldVolume.setEnabled(True)
                    self.sldVolume.setValue(camera.volume)
                    if camera.mute:
                        self.btnMute.setStyleSheet(self.getButtonStyle("mute"))
                    else:
                        self.btnMute.setStyleSheet(self.getButtonStyle("audio"))
                else:
                    self.btnMute.setEnabled(False)
                    self.sldVolume.setEnabled(False)

                self.btnRecord.setEnabled(True)
                if camera.isRecording():
                    self.btnRecord.setStyleSheet(self.getButtonStyle("recording"))
                    record_always = player.systemTabSettings().record_always if player.systemTabSettings() else False
                    record_alarm = player.systemTabSettings().record_alarm if player.systemTabSettings() else False
                    record_enable = player.systemTabSettings().record_enable if player.systemTabSettings() else False
                    if record_enable and ((camera.isAlarming() and record_alarm) or record_always):
                        self.btnRecord.setEnabled(False)
                else:
                    self.btnRecord.setStyleSheet(self.getButtonStyle("record"))
            else:
                reconnecting = False
                #self.tabVideo.btnSnapshot.setEnabled(False)
                timers = self.mw.pm.getStreamPairTimers(camera.uri())
                for timer in timers:
                    if timer.isActive():
                        reconnecting = True
                
                if reconnecting:
                    self.btnStop.setStyleSheet(self.getButtonStyle("stop"))
                else:
                    self.btnStop.setStyleSheet(self.getButtonStyle("play"))
                    self.setTabsEnabled(True)

                if profile := camera.getProfile(camera.uri()):
                    if camera.mute:
                        self.btnMute.setStyleSheet(self.getButtonStyle("mute"))
                    else:
                        self.btnMute.setStyleSheet(self.getButtonStyle("audio"))

                    if profile.audio_bitrate() and not profile.getDisableAudio():
                        self.btnMute.setEnabled(True)
                        self.sldVolume.setEnabled(True)
                    else:
                        self.btnMute.setEnabled(False)
                        self.sldVolume.setEnabled(False)

                self.btnRecord.setStyleSheet(self.getButtonStyle("record"))
                self.btnRecord.setEnabled(False)
                self.btnSnapshot.setEnabled(False)
        else:
            self.sldVolume.setEnabled(False)
            self.btnMute.setStyleSheet(self.getButtonStyle("audio"))
            self.btnMute.setEnabled(False)
            self.btnRecord.setStyleSheet(self.getButtonStyle("record"))
            self.btnRecord.setEnabled(False)
            self.btnStop.setStyleSheet(self.getButtonStyle("play"))
            self.btnStop.setEnabled(False)
            self.btnSnapshot.setEnabled(False)

        if self.lstCamera.count():
            self.btnStopAll.setEnabled(True) 
        else:
            self.btnStopAll.setEnabled(False)
        if len(self.mw.pm.players):
            self.btnStopAll.setStyleSheet(self.getButtonStyle("stop_all"))
        else:
            self.btnStopAll.setStyleSheet(self.getButtonStyle("play_all"))

        self.sync_lock = False


    def getButtonStyle(self, name):
        strStyle = "QPushButton { image : url(image:%1.png); } \
                    QPushButton:hover { image : url(image:%1_hi.png); } \
                    QPushButton:pressed { image : url(image:%1_lo.png); } \
                    QPushButton:disabled { image : url(image:%1_lo.png); }"
        strStyle = strStyle.replace("%1", name)
        return strStyle

    def getCurrentPlayer(self):
        result = None
        if self.lstCamera:
            camera = self.getCurrentCamera()
            if camera:
                result = self.mw.pm.getPlayer(camera.uri())
        return result

    def getCamera(self, uri):
        if not uri: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            for profile in camera.profiles:
                if profile.uri() == uri:
                    return camera

        return None
    
    def getCameraByName(self, name):
        if not name: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if camera.name() == name:
                return camera
                
        return None

    
    def getCameraBySerialNumber(self, serial_number):
        if not serial_number: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if camera.serial_number() == serial_number:
                return camera
                
        return None
    
    def getCameraByXAddrs(self, xaddrs):
        if not xaddrs: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if camera.xaddrs() == xaddrs:
                return camera
            
        return None
    
    def getProfile(self, uri):
        result = None
        if camera := self.getCamera(uri):
            result = camera.getProfile(uri)
        return result
    
    def getCurrentProfile(self):
        result = None
        if camera := self.getCurrentCamera():
            result = camera.getProfile(camera.uri())
        return result
    
    def getCurrentCamera(self):
        result = None
        if self.lstCamera:
            result = self.lstCamera.currentItem()
        return result
    
    def setCurrentCamera(self, uri):
        if camera := self.getCamera(uri):
            self.lstCamera.setCurrentItem(camera)
            self.signals.fill.emit(camera.onvif_data)
            self.syncGUI()

            if self.mw.videoConfigure:
                    self.mw.videoConfigure.setCamera(camera)

            if self.mw.audioConfigure:
                self.mw.audioConfigure.setCamera(camera)

    def enableAutoTimeSync(self, state):
        AUTO_TIME_SYNC_INTERVAL = 3600000
        if int(state) == 0:
            logger.debug("Auto time sync has been turned off")
            if self.autoTimeSyncer:
                self.autoTimeSyncer.stop()
        else:
            logger.debug("Auto time sync has been turned on")
            if not self.autoTimeSyncer:
                self.autoTimeSyncer = QTimer()
                self.autoTimeSyncer.setInterval(AUTO_TIME_SYNC_INTERVAL)
                self.autoTimeSyncer.timeout.connect(self.timeSync)
                self.autoTimeSyncer.start()
    
    def timeSync(self):
        if not self.lstCamera: return
        logger.debug("Synchronizing camera times")
        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            camera.onvif_data.startUpdateTime()

    def activeSessions(self):
        for session in self.sessions:
            if session.active:
                return True
            
        return False

    def closeEvent(self):
        self.closing = True

        if self.lstCamera:
            cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
            for camera in cameras:
                camera.onvif_data.filled = None

        if self.activeSessions():
            for session in self.sessions:
                session.abort = True

            sleep(1)
            
            waiting = True
            count = 0
            while waiting:
                count += 1
                tmp = False
                for session in self.sessions:
                    if session.active:
                        tmp = True
                        break
                    sleep(0.1)
                waiting = tmp
                if count > 10:
                    break

    def allCamerasFilled(self):
        if not self.lstCamera: return True

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if not camera.filled:
                return False

        return True


#/********************************************************************
# Cayenue/cayenue/panels/picture/picturepanel.py 
#
# Copyright (c) 2026  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
from loguru import logger
from PyQt6.QtWidgets import QGridLayout, QWidget, \
    QLabel, QAbstractItemView, QAbstractItemView, \
    QApplication, QTreeView, QMenu, QMessageBox
from PyQt6.QtGui import QFileSystemModel, QAction, QColorConstants
from PyQt6.QtCore import Qt, QFileInfo, QDateTime, \
    QDate, QTime, QStandardPaths
from cayenue.components.directoryselector import DirectorySelector
from cayenue.panels.file import FileControlPanel
from cayenue.components import Progress, InfoDialog
from pathlib import Path
import traceback

class PicTreeView(QTreeView):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw

    def keyPressEvent(self, event):
        pass_through = True

        match event.key():
            case Qt.Key.Key_Return:
                pass_through = False
                index = self.currentIndex()
                if index.isValid():
                    fileInfo = self.model().fileInfo(index)
                    if fileInfo.isFile():
                        self.mw.picturePanel.playVideo()
                    else:
                        if self.isExpanded(index):
                            self.collapse(index)
                        else:
                            self.expand(index)
            case Qt.Key.Key_Escape:
                if player := self.mw.filePanel.getCurrentlyPlayingFile():
                    self.mw.closeAllStreams()
                    pass_through = False
                else:
                    self.mw.glWidget.buffer.fill(QColorConstants.Black)
                    self.mw.glWidget.update()
            case Qt.Key.Key_Space:
                if player := self.mw.filePanel.getCurrentlyPlayingFile():
                    player.togglePaused()
            case Qt.Key.Key_Left:
                self.mw.filePanel.rewind()
                pass_through = False
            case Qt.Key.Key_Right:
                self.mw.filePanel.fastForward()
                pass_through = False
            case Qt.Key.Key_F1:
                self.mw.picturePanel.onMenuInfo()
            case Qt.Key.Key_F2:
                self.mw.picturePanel.onMenuRename()
            case Qt.Key.Key_Delete:
                self.mw.picturePanel.onMenuRemove()

        if pass_through:
            return super().keyPressEvent(event)
        
    def showCurrentFile(self):
        try:
            if self.mw.settings_profile != "Reader":
                return
            index = self.currentIndex()
            if index.isValid():
                info = self.model().fileInfo(index)
                if info.isFile() and self.model().isReadOnly():
                    self.mw.glWidget.drawFile(info.absoluteFilePath())
        except Exception as ex:
            logger.error(f"PicturePanel showCurrentFile exception: {ex}")

    def currentChanged(self, newIndex, oldIndex):
        if newIndex.isValid():
            self.showCurrentFile()
            self.scrollTo(newIndex)

class PicturePanel(QWidget):                
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw
        self.geometryKey = "PictureBrowseDialog/geometry"
        self.splitKey = "PictureBrowseDialog/splitSettings"
        self.headerKey = "PictureBrowseDialog/header"
        self.positionInitialized = False
        self.setWindowTitle("Event Browser")
        self.expandedPaths = []
        self.loadedCount = 0
        self.restorationPath = None
        self.verticalScrollBarPosition = 0
        self.control = FileControlPanel(mw, self)
        self.progress = Progress(mw)
        self.dlgInfo = InfoDialog(mw)

        if self.mw.parent_window:
            picture_dirs = self.mw.parent_window.settingsPanel.storage.dirPictures.text()
        else:
            picture_dirs = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.PicturesLocation)[0]
        self.dirPictures = DirectorySelector(mw, self.mw.settingsPanel.storage.pictureKey, "", picture_dirs)
        #self.dirPictures.signals.dirChanged.connect(self.mw.settingsPanel.storage.dirPicturesChanged)
        self.dirPictures.signals.dirChanged.connect(self.dirChanged)

        self.model = QFileSystemModel(mw)
        self.model.directoryLoaded.connect(self.loaded)
        self.tree = PicTreeView(mw)
        self.tree.setModel(self.model)
        #self.tree.setSortingEnabled(True)
        #self.tree.sortByColumn(0, Qt.SortOrder.DescendingOrder)
        self.tree.doubleClicked.connect(self.treeDoubleClicked)
        self.tree.setColumnHidden(1, True)
        self.tree.setColumnHidden(2, True)
        if data := self.mw.settings.value(self.headerKey):
            self.tree.header().restoreState(data)
        self.tree.update()
        self.tree.header().sectionResized.connect(self.headerChanged)
        self.tree.header().sectionMoved.connect(self.headerChanged)
        self.tree.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.showContextMenu)

        pnlTree = QWidget()
        lytTree = QGridLayout(pnlTree)
        lytTree.addWidget(self.dirPictures,       0, 0, 1, 1)
        lytTree.addWidget(self.tree,              1, 0, 1, 1)
        lytTree.addWidget(self.progress,          2, 0, 1, 1)
        lytTree.addWidget(QLabel(),               3, 0, 1, 1)
        lytTree.addWidget(self.control,  4, 0, 1, 1)

        lytMain = QGridLayout(self)
        lytMain.addWidget(pnlTree)
        lytMain.setContentsMargins(0, 0, 0, 0)

        self.dirChanged(self.dirPictures.text())

        self.menu = QMenu("Context Menu", self)
        self.remove = QAction("Delete", self)
        self.rename = QAction("Rename", self)
        self.info = QAction("Info", self)
        self.remove.triggered.connect(self.onMenuRemove)
        self.rename.triggered.connect(self.onMenuRename)
        self.info.triggered.connect(self.onMenuInfo)
        self.menu.addAction(self.remove)
        self.menu.addAction(self.rename)
        self.menu.addAction(self.info)

    def headerChanged(self, a, b, c):
        self.mw.settings.setValue(self.headerKey, self.tree.header().saveState())

    def closeEvent(self, event):
        self.mw.settings.setValue(self.geometryKey, self.geometry())
        return super().closeEvent(event)

    def showEvent(self, event):
        if rect := self.mw.settings.value(self.geometryKey):
            if rect.width() and rect.height():
                self.setGeometry(rect)
                self.positionInitialized = True
        return super().showEvent(event)
    
    def splitterMoved(self, pos, index):
        self.mw.settings.setValue(self.splitKey, self.split.saveState())

    def filenameToQDateTime(self, filename):
        # file start times may be inaccurate on samba, use filename instead
        filename = Path(filename).stem
        year = int(filename[0:4])
        month = int(filename[4:6])
        date = int(filename[6:8])
        hour = int(filename[8:10])
        minute = int(filename[10:12])
        second = int(filename[12:14])
        qdate = QDate(year, month, date)
        qtime = (QTime(hour, minute, second))
        return QDateTime(qdate, qtime)
    
    def searchFiles(self, target, dir):
        try:
            alarm_buffer_size = self.mw.settingsPanel.alarm.spnBufferSize.value()
            files = os.listdir(dir)
            for file in files:
                file_info = QFileInfo(str(Path(dir) / file))
                start = self.filenameToQDateTime(file).addSecs(-alarm_buffer_size)
                finish = file_info.lastModified()
                if (target >= start and target <= finish):
                    return file
            return None
        except Exception as ex:
            logger.error(f"PicturePanel searchFile exception: {ex}")

    def calculateTargetPct(self, target, file, dir):
        alarm_buffer_size = self.mw.settingsPanel.alarm.spnBufferSize.value()
        start = self.filenameToQDateTime(file).addSecs(-alarm_buffer_size)
        finish = QFileInfo(str(Path(dir) / file)).lastModified()
        numerator = start.secsTo(target)
        denominator = start.secsTo(finish)
        if denominator > 0:
            pct = float(numerator / denominator)
            if pct > 0.98:
                pct = 0.95
        else:
            pct = 0.95
        return pct

    def playVideo(self):
        try:
            alarm_buffer_size = self.mw.settingsPanel.alarm.spnBufferSize.value()
            index = self.tree.currentIndex()
            if index.isValid():
                pic_info = self.model.fileInfo(index)
                if pic_info.isFile():
                    target = self.filenameToQDateTime(pic_info.fileName())
                    dir = Path(self.mw.filePanel.dirArchive.txtDirectory.text()) / str(pic_info.absoluteDir().dirName())

                    # everything is approximate, so give the algorithm a couple chances to find near miss
                    found = False
                    if file := self.searchFiles(target, dir):
                        found = True
                    else:
                        fuzz = target.addSecs(-alarm_buffer_size)
                        if file := self.searchFiles(fuzz, dir):
                            target = fuzz
                            found = True
                        else:
                            fuzz = target.addSecs(alarm_buffer_size)
                            if file := self.searchFiles(fuzz, dir):
                                target = fuzz
                                found = True

                    if not found:
                        return
                    
                    self.selectFileInFilePanelTree(str(dir), file)
                    pct = self.calculateTargetPct(target, file, dir)
                    vid_info = QFileInfo(os.path.join(dir, file))
                    if player := self.mw.pm.getPlayer(vid_info.absoluteFilePath()):
                        player.seek(pct)
                    else:
                        self.mw.closeAllStreams()
                        self.mw.playMedia(vid_info.absoluteFilePath(), file_start_from_seek=pct)
                        self.mw.glWidget.focused_uri = vid_info.absoluteFilePath()

        except Exception as ex:
            logger.error(f'Event browser exception: {ex}')
            logger.debug(traceback.format_exc())

    def selectFileInFilePanelTree(self, path, filename):
        tree = self.mw.filePanel.tree
        model = tree.model()
        if camera_idx := model.index(path):
            if camera_idx.isValid():
                if not tree.isExpanded(camera_idx):
                    tree.setExpanded(camera_idx, True)
                if file_idx := model.index(os.path.join(path, filename)):
                    if file_idx.isValid():
                        tree.setCurrentIndex(file_idx)
                        tree.scrollTo(file_idx, QAbstractItemView.ScrollHint.PositionAtCenter)

    def loaded(self, path):
        self.loadedCount += 1
        self.model.sort(0)
        #self.model.sort(0, Qt.SortOrder.DescendingOrder)
        for i in range(self.model.rowCount(self.model.index(path))):
            idx = self.model.index(i, 0, self.model.index(path))
            if idx.isValid():
                if self.model.filePath(idx) in self.expandedPaths:
                    self.tree.setExpanded(idx, True)

        if len(self.expandedPaths):
            if self.loadedCount == len(self.expandedPaths) + 1:
                if self.verticalScrollBarPosition:
                    QApplication.processEvents()
                    self.tree.verticalScrollBar().setValue(self.verticalScrollBarPosition)
                self.expandedPaths.clear()
                self.restoreSelectedPath()
        else:
            self.restoreSelectedPath()

    def restoreSelectedPath(self):
        if self.restorationPath:
            idx = self.model.index(self.restorationPath)
            if idx.isValid():
                self.tree.setCurrentIndex(idx)
            self.restorationPath = None

    def refresh(self):
        try:
            self.loadedCount = 0
            self.expandedPaths = []
            self.restorationPath = self.model.filePath(self.tree.currentIndex())
            path = self.dirPictures.txtDirectory.text()
            self.model.sort(0)
            for i in range(self.model.rowCount(self.model.index(path))):
                idx = self.model.index(i, 0, self.model.index(path))
                if idx.isValid():
                    if self.tree.isExpanded(idx):
                        self.expandedPaths.append(self.model.filePath(idx))
            self.verticalScrollBarPosition = self.tree.verticalScrollBar().value()
            self.model = QFileSystemModel()
            self.model.setRootPath(path)
            self.model.directoryLoaded.connect(self.loaded)
            self.tree.setModel(self.model)
            self.tree.setRootIndex(self.model.index(path))
        except Exception as ex:
            logger.error(f"PicturePanel refresh exception: {ex}")

    def dirChanged(self, path):
        if len(path) > 0:
            self.model.setRootPath(path)
            self.tree.setRootIndex(self.model.index(path))

    def treeDoubleClicked(self, index):
        if index.isValid():
            fileInfo = self.model.fileInfo(index)
            if fileInfo.isDir():
                self.tree.setExpanded(index, self.tree.isExpanded(index))
            else:
                self.playVideo()

    # some weird bug makes this not visible to calling signals in some cases
    def onMediaStarted(self, duration):
        if self.mw.tab.currentIndex() == 1:
            self.tree.setFocus()
        self.control.setBtnPlay()
        self.control.setBtnMute()
        self.control.setSldVolume()

    def onMediaStopped(self, uri):
        self.control.setBtnPlay()
        self.progress.updateProgress(0.0)
        self.progress.lblDuration.setText("0:00")
        # not relavent to camera playback
        self.tree.showCurrentFile()

    def onMediaProgress(self, pct, uri):
        if player := self.mw.pm.getPlayer(uri):
            player.file_progress = pct
            self.progress.updateDuration(player.duration())

        #if pct >= 0.0 and pct <= 1.0 and uri == self.mw.glWidget.focused_uri:
        if pct >= 0.0 and pct <= 1.0:
            self.progress.updateProgress(pct)

    def showContextMenu(self, pos):
        index = self.tree.indexAt(pos)
        if index.isValid():
            fileInfo = self.model.fileInfo(index)
            if fileInfo.isFile():
                self.menu.exec(self.mapToGlobal(pos))

    def onMenuRemove(self):
        index = self.tree.currentIndex()
        if index.isValid():
            ret = QMessageBox.warning(self, "Cayenue",
                                        "You are about to ** PERMANENTLY ** delete this file.\n"
                                        "Are you sure you want to continue?",
                                        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)

            if ret == QMessageBox.StandardButton.Ok:
                try:
                    idxAbove = self.tree.indexAbove(index)
                    idxBelow = self.tree.indexBelow(index)

                    self.model.remove(index)
                    
                    resolved = False
                    if idxAbove.isValid():
                        if os.path.isfile(self.model.filePath(idxAbove)):
                            self.tree.setCurrentIndex(idxAbove)
                            resolved = True
                    if not resolved:
                        if idxBelow.isValid():
                            if os.path.isfile(self.model.filePath(idxBelow)):
                                self.tree.setCurrentIndex(idxBelow)
                
                except Exception as e:
                    logger.error(f'File delete error: {e}')

    def onMenuRename(self):
        index = self.tree.currentIndex()
        if index.isValid():
            self.model.setReadOnly(False)
            self.tree.edit(index)

    def onFileRenamed(self, path, oldName, newName):
        self.model.setReadOnly(True)

    def onMenuInfo(self):
        strInfo = ""
        try:
            index = self.tree.currentIndex()
            if (index.isValid()):
                info = self.model.fileInfo(index)
                strInfo += f"Filename:  {info.fileName()}"
                strInfo += f"\nCreated:  {info.birthTime().toString()}"
                width = self.mw.glWidget.pixmap.width()
                height = self.mw.glWidget.pixmap.height()
                strInfo += f"\nImage Dims: {width} x {height}"           
            else:
                strInfo = "Invalid Index"
        except Exception as ex:
            strInfo = f'Unable to read file info: {ex}'

        #msgBox = QMessageBox(self)
        #msgBox.setWindowTitle("File Info")
        #msgBox.setText(strInfo)
        #msgBox.exec()
        self.dlgInfo.lblMessage.setText(strInfo)
        self.dlgInfo.exec()



import http.server
import socketserver
import os
import signal
import sys
from pathlib import Path

PORT = 8800
if sys.platform == "win32":
    DIRECTORY = f"{os.environ['HOMEPATH']}/.cache/Cayenue/proxy"
else:
    DIRECTORY = f"{os.environ['HOME']}/.cache/Cayenue/proxy"

if sys.platform == "win32":
    DIRECTORY = f"{os.environ['HOMEPATH']}/.cache/Cayenue/proxy"
elif sys.platform == "darwin":
    DIRECTORY = Path(os.path.dirname(__file__)).parent.absolute() / "cache" / "proxy"
elif sys.platform.startswith("linux"):
    DIRECTORY = f"{os.environ['HOME']}/.cache/Cayenue/proxy"
else:
    print(f"Unknown platform: {sys.platform}")

def handle_sigterm(signum, frame):
    sys.exit(0)

signal.signal(signal.SIGTERM, handle_sigterm)

class Server(socketserver.ThreadingTCPServer):
    allow_reuse_address = True
    disable_nagle_algorithm = True

class Handler(http.server.SimpleHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=DIRECTORY, **kwargs)

    def do_POST(self):
        if self.path == "/shutdown":
            self.send_response(200)
            self.end_headers()
            self.server.shutdown()
        else:
            super().do_POST()

if __name__ == "__main__":
    try:
        with Server(("", PORT), Handler) as httpd:
            httpd.serve_forever()
    except Exception as ex:
        print(f"HTTP SERVER ERROR: {ex}")




#********************************************************************
# Cayenue/cayenue/components/target.py
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QDialog, QGridLayout, QListWidget, QListWidgetItem, \
    QDialogButtonBox, QWidget, QLabel, QPushButton, QMessageBox, QSlider, QCheckBox
from PyQt6.QtCore import Qt, pyqtSignal, QObject
from .warningbar import WarningBar, Indicator
from loguru import logger

class Target(QListWidgetItem):
    def __init__(self, name, id):
        super().__init__(name)
        self.id = id

class TargetDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw

        self.targets = {
            0: "person",
            1: "bicycle",
            2: "car",
            3: "motorcycle",
            4: "airplane",
            5: "bus",
            6: "train",
            7: "truck",
            8: "boat",
            14: "bird",
            15: "cat",
            16: "dog",
            17: "horse",
            18: "sheep",
            19: "cow",
            21: "bear"
        }

        self.setModal(True)
        self.setWindowTitle("Add Target")
        self.list = QListWidget()
        for key in self.targets:
            self.list.addItem(Target(self.targets[key], key))

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Close)
        self.buttonBox.button(QDialogButtonBox.StandardButton.Close).setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.buttonBox.rejected.connect(self.reject)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.list,       0, 0, 1, 2)
        lytMain.addWidget(self.buttonBox,  1, 0, 1, 2)

        self.list.setCurrentRow(0)

    def reject(self):
        self.hide()

class TargetListSignals(QObject):
    delete = pyqtSignal()

class TargetList(QListWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.signals = TargetListSignals()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Delete:
            self.signals.delete.emit()
        return super().keyPressEvent(event)
    
    def toString(self):
        output = ""
        length = self.count()
        for i in range(length):
            output += str(self.item(i).id)
            if i < length - 1:
                output += ":"
        return output

class TargetSelector(QWidget):
    def __init__(self, mw, module):
        super().__init__()
        self.mw = mw
        self.module = module

        self.lstTargets = TargetList(self.mw)
        self.lstTargets.signals.delete.connect(self.btnDeleteTargetClicked)
        self.lblTargets = QLabel("Targets")
        self.btnAddTarget = QPushButton("+")
        self.btnAddTarget.clicked.connect(self.btnAddTargetClicked)
        self.btnDeleteTarget = QPushButton("-")
        self.btnDeleteTarget.clicked.connect(self.btnDeleteTargetClicked)
        self.dlgTarget = TargetDialog(self.mw)
        self.dlgTarget.list.itemDoubleClicked.connect(self.onAddItemDoubleClicked)
        self.dlgTarget.buttonBox.accepted.connect(self.dlgListAccepted)

        # gui works better if these are on the same panel
        self.barLevel = WarningBar()
        self.indAlarm = Indicator(self.mw)
        self.sldGain = QSlider(Qt.Orientation.Vertical)
        self.sldGain.setMinimum(0)
        self.sldGain.setMaximum(100)
        self.sldGain.setValue(0)
        self.sldGain.valueChanged.connect(self.sldGainValueChanged)
        self.lblGain = QLabel("0")

        pnlTargets = QWidget()
        pnlTargets.setMaximumWidth(200)
        lytTargets = QGridLayout(pnlTargets)

        self.chkShowBoxes = QCheckBox("Show Boxes")
        self.chkShowBoxes.stateChanged.connect(self.chkShowBoxesStateChanged)

        lytTargets.addWidget(self.lblTargets,       0, 0, 1, 1)
        lytTargets.addWidget(self.btnDeleteTarget,  0, 1, 1, 1)
        lytTargets.addWidget(self.btnAddTarget,     0, 2, 1, 1)
        lytTargets.addWidget(self.lstTargets,       1, 0, 2, 3)

        lytMain = QGridLayout(self)
        lytMain.addWidget(pnlTargets,            1, 0, 2, 1)
        lytMain.addWidget(self.lblGain,          1, 1, 1, 1, Qt.AlignmentFlag.AlignHCenter)
        lytMain.addWidget(self.sldGain,          2, 1, 1, 1, Qt.AlignmentFlag.AlignHCenter)
        lytMain.addWidget(QLabel("Limit"),       3, 1, 1, 1, Qt.AlignmentFlag.AlignHCenter)
        lytMain.addWidget(self.indAlarm,         1, 2, 1, 1)
        lytMain.addWidget(self.barLevel,         2, 2, 1, 1)
        lytMain.addWidget(QLabel(),              2, 3, 1, 1)
        lytMain.addWidget(self.chkShowBoxes,     3, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytMain.setContentsMargins(0, 0, 0, 0)
        lytMain.setColumnStretch(2, 10)

    def btnAddTargetClicked(self):
        self.dlgTarget.show()

    def btnDeleteTargetClicked(self):
        try:
            if item := self.lstTargets.currentItem():
                ret = QMessageBox.warning(self, "Delete Target: " + item.text(), "You are about to delete target\n"
                                        "Are you sure you want to continue?",
                                        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
                if ret != QMessageBox.StandardButton.Ok:
                    return
                self.lstTargets.takeItem(self.lstTargets.row(item))

                if not self.mw.videoConfigure:
                    return
                
                if camera := self.mw.cameraPanel.getCurrentCamera():
                    if camera.videoModelSettings:
                        camera.videoModelSettings.setTargets(self.lstTargets.toString())

        except Exception as ex:
            logger.error(f'Error deleting target : {ex}')

    def dlgListAccepted(self):
        item = self.dlgTarget.list.item(self.dlgTarget.list.currentRow())
        self.onAddItemDoubleClicked(item)

    def onAddItemDoubleClicked(self, item):
        try:
            target = Target(item.text(), item.id)
            found = False
            for i in range(self.lstTargets.count()):
                if target.text() == self.lstTargets.item(i).text():
                    found = True
                    break
            if not found:
                self.lstTargets.addItem(target)
                if not self.mw.videoConfigure:
                    return
                if camera := self.mw.cameraPanel.getCurrentCamera():
                    if camera.videoModelSettings:
                        camera.videoModelSettings.setTargets(self.lstTargets.toString())

        except Exception as ex:
            logger.error(f'Target add itme error : {ex}')

    def setTargets(self, targets):
        while self.lstTargets.count() > 0:
            self.lstTargets.takeItem(0)

        for t in targets:
            self.lstTargets.addItem(Target(self.dlgTarget.targets[t], t))

    def getTargets(self):
        output = []
        for i in range(self.lstTargets.count()):
            target = self.lstTargets.item(i)
            output.append(target.id)
        return output
    
    def setModelParameters(self, videoModelSettings):
        self.setTargets(videoModelSettings.targets)
        self.disconnectSlider()
        self.sldGain.setMaximum(videoModelSettings.sampleSize)
        self.sldGain.setValue(videoModelSettings.limit)
        self.lblGain.setText(str(videoModelSettings.limit))
        self.connectSlider()
        self.chkShowBoxes.setChecked(videoModelSettings.show)
        self.barLevel.setLevel(0)
        self.indAlarm.setState(0)

    def disconnectSlider(self):
        self.sldGain.valueChanged.disconnect()

    def connectSlider(self):
        self.sldGain.valueChanged.connect(self.sldGainValueChanged)

    def sldGainValueChanged(self, value):
        try:
            self.lblGain.setText(f'{value}')
            if not self.mw.videoConfigure:
                return

            if camera := self.mw.cameraPanel.getCurrentCamera():
                if camera.videoModelSettings:
                    camera.videoModelSettings.setModelOutputLimit(value)

        except Exception as ex:
            logger.error(f'Error changing gain slider value : {ex}')

    def chkShowBoxesStateChanged(self, state):
        try:
            if not self.mw.videoConfigure:
                return

            if camera := self.mw.cameraPanel.getCurrentCamera():
                if camera.videoModelSettings:
                    camera.videoModelSettings.setModelShowBoxes(bool(state))

        except Exception as ex:
            logger.error(f'Error changing show boxes state : {ex}')



#/********************************************************************
# Cayenue/panels/video/modules/common/yolosettings.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

class VideoModelSettings():
    def __init__(self, mw, camera, module_name=None):
        self.camera = camera
        self.mw = mw
        self.module_name = module_name
        self.id = camera.serial_number()

        self.targets = self.getTargetsForPlayer()
        self.limit = self.getModelOutputLimit()
        self.confidence = self.getModelConfidence()
        self.show = self.getModelShowBoxes()
        self.skipFrames = self.getSkipFrames()
        self.skipCounter = 0
        self.sampleSize = self.getSampleSize()
        self.gain = self.getModelOutputGain()

    def getTargets(self):
        key = f'{self.id}/{self.module_name}/Targets'
        return str(self.mw.settings.value(key, "")).strip()
    
    def getTargetsForPlayer(self):
        var = self.getTargets()
        ary = []
        if len(var):
            tmp = var.split(":")
            for t in tmp:
                ary.append(int(t))
        return ary    

    def setTargets(self, targets):
        key = f'{self.id}/{self.module_name}/Targets'
        self.targets.clear()
        if len(targets):
            tmp = targets.split(":")
            for t in tmp:
                self.targets.append(int(t))
        self.mw.settings.setValue(key, targets)

    def getModelConfidence(self):
        key = f'{self.id}/{self.module_name}/ConfidenceThreshold'
        return int(self.mw.settings.value(key, 50))
    
    def setModelConfidence(self, value):
        key = f'{self.id}/{self.module_name}/ConfidenceThreshold'
        self.confidence = value
        self.mw.settings.setValue(key, value)

    def getModelOutputLimit(self):
        key = f'{self.id}/{self.module_name}/ModelOutputLimit'
        return int(self.mw.settings.value(key, 0))
    
    def setModelOutputLimit(self, value):
        key = f'{self.id}/{self.module_name}/ModelOutputLimit'
        self.limit = value
        self.mw.settings.setValue(key, value)

    def getModelShowBoxes(self):
        key = f'{self.id}/{self.module_name}/ModelShowBoxes'
        return bool(int(self.mw.settings.value(key, 1)))
    
    def setModelShowBoxes(self, value):
        key = f'{self.id}/{self.module_name}/ModelShowBoxes'
        self.show = value
        self.mw.settings.setValue(key, int(value))

    def getSkipFrames(self):
        key = f'{self.id}/{self.module_name}/SkipFrames'
        return int(self.mw.settings.value(key, 0))

    def setSkipFrames(self, value):
        key = f'{self.id}/{self.module_name}/SkipFrames'
        self.skipFrames = int(value)
        self.mw.settings.setValue(key, int(value))

    def getSampleSize(self):
        key = f'{self.id}/{self.module_name}/SampleSize'
        return int(self.mw.settings.value(key, 1))

    def setSampleSize(self, value):
        key = f'{self.id}/{self.module_name}/SampleSize'
        self.sampleSize = int(value)
        self.mw.settings.setValue(key, int(value))        

    def getModelOutputGain(self):
        key = f'{self.id}/{self.module_name}/MotionGain'
        return int(self.mw.settings.value(key, 50))
    
    def setModelOutputGain(self, value):
        key = f'{self.id}/{self.module_name}/MotionGain'
        self.gain = value
        self.mw.settings.setValue(key, value)




#********************************************************************
# Cayenue/cayenue/glwidget.py
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtOpenGLWidgets import QOpenGLWidget
from PyQt6.QtGui import QPainter, QImage, QColorConstants, QPen, QMovie, QIcon, QPixmap
from PyQt6.QtCore import QSize, QPointF, QRectF, QTimer, Qt, QRect, QPoint
from PyQt6.QtWidgets import QMessageBox
import numpy as np
from datetime import datetime
import time
from cayenue.enums import StreamState, ProxyType
from loguru import logger
import threading
import traceback

class GLWidget(QOpenGLWidget):
    def __init__(self, mw):
        super().__init__()

        self.mw = mw
        self.focused_uri = None
        self.model_loading = False
        self.spinner = QMovie("image:spinner.gif")
        self.spinner.start()
        self.plain_recording = QMovie("image:plain_recording.gif")
        self.plain_recording.start()
        self.alarm_recording = QMovie("image:alarm_recording.gif")
        self.alarm_recording.start()

        self.buffer = QImage(self.size(), QImage.Format.Format_RGB888)
        self.file_draw = False
        self.pixmap = None
        self.force_clear = False

        self.timer = QTimer()
        self.timer.timeout.connect(self.timerCallback)
        refreshInterval = self.mw.settingsPanel.general.spnDisplayRefresh.value()
        self.timer.start(refreshInterval)

        self.last_alarm_check = time.time()
        self.alarms = {}
        self.last_player_count = 1
        self.last_manage_directory = datetime.now()
    
    def renderCallback(self, F, uri):
        try :
            if F is None:
                logger.error(f"a null value was returned for frame {uri}")
                return

            player = self.mw.pm.getPlayer(uri)
            if not player:
                logger.error(f"a null value was returned for player {uri}")
                return
            
            camera = self.mw.cameraPanel.getCamera(player.uri)
            if player.isCameraStream() and not camera:
                logger.error(f'a null value was returned for camera {uri}')
                return

            player.needs_render = True

            self.mw.pm.lock()
            player.lock()

            player.ary = np.array(F, copy=True)

            if len(player.ary.shape) < 3:
                return
            h = player.ary.shape[0]
            w = player.ary.shape[1]
            d = player.ary.shape[2]
            player.image = QImage(player.ary.data, w, h, d * w, QImage.Format.Format_RGB888)

            if self.mw.settingsPanel.proxy.generateAlarmsLocally():
                if self.mw.videoConfigure:
                    if player.analyze_video and self.mw.videoConfigure.initialized:
                        if camera.videoModelSettings.skipCounter < camera.videoModelSettings.skipFrames:
                            camera.videoModelSettings.skipCounter += 1
                        else:
                            camera.videoModelSettings.skipCounter = 0
                            player.boxes = self.mw.pyVideoCallback(player.ary, uri)
                            result = player.processModelOutput()
                            alarmState = result >= camera.videoModelSettings.limit if result else False
                            player.handleAlarm(alarmState)
                            if camera.isCurrent():
                                level = 0
                                if camera.videoModelSettings.limit:
                                    level = result / camera.videoModelSettings.limit
                                else:
                                    if result:
                                        level = 1.0

                                if camera.videoModelSettings.module_name == "yolox":
                                    self.mw.videoConfigure.selTargets.barLevel.setLevel(level)
                                    if alarmState:
                                        self.mw.videoConfigure.selTargets.indAlarm.setState(1)
                    else:
                        # clear the video panel alarm display
                        if player.uri == self.focused_uri:
                            if self.mw.videoWorker:
                                self.mw.videoWorker(None, None)
            else:
                player.loadRemoteDetections()


            w_s = w
            h_s = h
            # the aspect ratios are multiplied by 100 and converted to int for comparison purpose
            actual_ratio = int(100.0 * float(w) / float(h))
            if actual_ratio and player.desired_aspect:
                if actual_ratio != player.desired_aspect:
                    if player.desired_aspect > 1:
                        w_s = int(h * player.desired_aspect / 100)
                    else:
                        h_s = int(w * 100 / player.desired_aspect)

            self.mw.pm.sizes[player.uri] = QSize(w_s, h_s)


            if player.packet_drop_frame_counter > 0:
                player.packet_drop_frame_counter -= 1
            else:
                player.packet_drop_frame_counter = 0

        except BaseException as ex:
            logger.error(f'GLWidget render callback exception: {str(ex)}')
            logger.debug(traceback.format_exc())

        player.unlock()
        self.mw.pm.unlock()
        self.alarmBroadcast()

    def reconnectCycle(self):
        self.mw.pm.lock()
        try:
            for player in self.mw.pm.players.values():
                if not player.last_render:
                    player.last_render = datetime.now()
                if player.isCameraStream() and player.last_render:
                    interval = datetime.now() - player.last_render
                    if interval.total_seconds() > 30:
                        player.requestShutdown(reconnect=True)
                        player.last_render = datetime.now()
                        continue
        except Exception as ex:
            logger.error(f'GLWidget reconnect exception: {str(ex)}')
        self.mw.pm.unlock()

    def timerCallback(self):

        if self.force_clear:
            self.force_clear = False
            if self.mw.settings_profile != "Reader":
                self.buffer.fill(QColorConstants.Black)


        if (not self.mw.pm.countPlayers() and not self.mw.getActiveTimerCount() and not self.last_player_count):
            return

        # limit calls to manage directory to once per cycle to reduce load
        md_interval = datetime.now() - self.last_manage_directory
        if md_interval.total_seconds() > self.mw.STD_FILE_DURATION:
            d = self.mw.settingsPanel.storage.dirArchive.txtDirectory.text()
            if self.mw.settingsPanel.storage.chkManageDiskUsage.isChecked() and self.mw.enable_disk_mgmt:
                thread = threading.Thread(target=self.mw.diskManager.manageDirectory, args=(d,))
                thread.start()
                self.last_manage_directory = datetime.now()
        
        # stress testing
        if self.mw.settingsPanel.general.chkStressTest.isChecked():
            self.reconnectCycle()
        
        if self.mw.last_alarm:
            interval = datetime.now() - self.mw.last_alarm
            if interval.total_seconds() > 10:
                self.mw.alarm_states = []

        if self.mw.split.sizes()[0]:
            if self.mw.pm.countPlayers() or len(self.mw.timers):
                self.file_draw = False
                self.buildImage()
            else:
                if self.buffer and not self.file_draw:
                    self.buffer.fill(QColorConstants.Black)

            self.update()

        self.last_player_count = self.mw.pm.countPlayers() + self.mw.getActiveTimerCount()

    def paintGL(self):
        try:
            if self.buffer:
                if not self.buffer.isNull():
                    painter = QPainter(self)
                    painter.setRenderHint(QPainter.RenderHint.Antialiasing)
                    painter.drawImage(self.rect(), self.buffer)
        except Exception as ex:
            logger.error(f'GLWidget paintGL exception: {str(ex)}')
            logger.debug(traceback.format_exc())


    def alarmBroadcast(self):
        proxyPanel = self.mw.settingsPanel.proxy
        if proxyPanel.proxyType == ProxyType.SERVER and proxyPanel.grpAlarmBroadcast.isChecked():
            interval = time.time() - self.last_alarm_check
            if interval > 1:
                self.last_alarm_check = time.time()
                camera_alarms = {}
                alarm_states = []
                if self.mw.cameraPanel:
                    if self.mw.cameraPanel.lstCamera:
                        lstCamera = self.mw.cameraPanel.lstCamera
                        cameras = [lstCamera.item(x) for x in range(lstCamera.count())]
                        for camera in cameras:
                            camera_alarms[camera.serial_number()] = camera.isAlarming()
                            alarm_states.append(camera.isAlarming())
                
                udp_msg = str(datetime.now()) + "\n\nALARMS"
                tmp = []
                self.alarms = camera_alarms
                for alarm_state in alarm_states:
                    tmp.append(alarm_state)
                    udp_msg += "\n\n" + str(int(alarm_state))

                self.mw.broadcaster.send(udp_msg)

    def sizeHint(self):
        return QSize(640, 480)

    def resizeEvent(self, event):
        if self.buffer.isNull():
            self.buffer = QImage(self.size(), QImage.Format.Format_RGB888)
        self.buffer = self.buffer.scaled(event.size())
        if self.file_draw:
            self.drawFile("")
        return super().resizeEvent(event)

    def mouseDoubleClickEvent(self, event):
        if self.mw.settingsPanel.proxy.proxyType == ProxyType.STAND_ALONE:
            ret = QMessageBox.warning(self, "Incompatible Configuration",
                                            "Focus Window is not available in stand alone configuration, please use Settings -> Proxy Server mode to enable",
                                            QMessageBox.StandardButton.Ok)
            return
        
        if self.mw.settings_profile == "Focus":
            return super().mouseDoubleClickEvent(event)

        if camera := self.mw.cameraPanel.getCurrentCamera():
            if profile := camera.getRecordProfile():
                self.mw.openFocusWindow()
                count = 0
                while not self.mw.focus_window.cameraPanel.getCamera(profile.uri()):
                    time.sleep(0.01)
                    count += 1
                    if count > 200:
                        logger.error("timeout error opening focus window")
                        break
                if camera := self.mw.focus_window.cameraPanel.getCamera(profile.uri()):
                    self.mw.focus_window.cameraPanel.onItemDoubleClicked(camera)

        return super().mouseDoubleClickEvent(event)
    
    def mousePressEvent(self, event):
        resolved = False
        self.mw.pm.lock()
        for player in self.mw.pm.players.values():
            if self.mw.pm.displayRect(player.uri, self.buffer.size()).contains(event.position()):
                if not player.hidden:
                    self.focused_uri = player.uri
                    if self.mw.isCameraStreamURI(player.uri):
                        self.mw.cameraPanel.setCurrentCamera(player.uri)
                        resolved = True
                        break
                    else:
                        if self.mw.isSplitterCollapsed():
                            self.mw.restoreSplitter()
                        self.mw.filePanel.setCurrentFile(self.focused_uri)
                        resolved = True
                        break
        self.mw.pm.unlock()
        if not resolved:
            for timer in self.mw.timers.values():
                if timer.isActive():
                    if self.mw.pm.displayRect(timer.uri, self.buffer.size()).contains(event.position()):
                        self.focused_uri = timer.uri
                        if self.mw.isSplitterCollapsed():
                            self.mw.restoreSplitter()
                        self.mw.cameraPanel.setCurrentCamera(self.focused_uri)
                        break
        return super().mousePressEvent(event)
    
    def isFocusedURI(self, uri):
        result = False
        if uri == self.focused_uri:
            result = True
        else:
            uris = self.mw.pm.getStreamPairURIs(uri)
            for u in uris:
                if u == self.focused_uri:
                    result = True
        return result

    def drawFile(self, filename):
        if self.buffer.isNull():
            return
        
        painter = QPainter(self.buffer)
        if not painter.isActive():
            return

        if self.mw.pm.countPlayers():
            return

        if len(filename):
            self.pixmap = QPixmap(filename)
        scaled_size = self.pixmap.size().scaled(self.size(), Qt.AspectRatioMode.KeepAspectRatio)
        scaled_rect = QRect(QPoint(0, 0), scaled_size)
        painter.fillRect(self.rect(), QColorConstants.Black)
        scaled_rect.moveCenter(self.rect().center())
        painter.drawPixmap(scaled_rect, self.pixmap)
        self.file_draw = True
        self.update()
    
    def buildImage(self):
        try:
            if self.buffer.isNull():
                return
            
            painter = QPainter(self.buffer)
            if not painter.isActive():
                return
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)

            if self.model_loading:
                rectSpinner = QRectF(0, 0, 40, 40)
                rectSpinner.moveCenter(QPointF(self.rect().center()))
                painter.drawImage(rectSpinner, self.spinner.currentImage())
                return
            
            if not len(self.mw.pm.players):
                painter.fillRect(self.rect(), QColorConstants.Black)

            self.mw.pm.lock()
            for player in self.mw.pm.players.values():

                if player.isRecording():
                    self.mw.enable_disk_mgmt = True

                if player.output_file_start_time:
                    interval = datetime.now() - player.output_file_start_time
                    if interval.total_seconds() > self.mw.STD_FILE_DURATION:
                        if filename := player.getOutputFilename():
                            player.startFileBreak(filename)

                if player.disable_video or player.hidden:
                    continue

                if not player.image and player.isCameraStream():
                    rect = self.mw.pm.displayRect(player.uri, self.buffer.size())
                    if rect.isValid():
                        painter.fillRect(rect, QColorConstants.Black)
                        rectSpinner = QRectF(0, 0, 40, 40)
                        rectSpinner.moveCenter(rect.center())
                        painter.drawImage(rectSpinner, self.spinner.currentImage())
                        if self.isFocusedURI(player.uri):
                            painter.setPen(QColorConstants.White)
                            painter.drawRect(rect.adjusted(1, 1, -2, -2))
                    if camera := self.mw.cameraPanel.getCamera(player.uri):
                        camera.setIcon(QIcon(self.spinner.currentPixmap()))
                        if camera.isCurrent():
                            self.mw.cameraPanel.setTabsEnabled(False)
                    continue

                if camera := self.mw.cameraPanel.getCamera(player.uri):
                    camera.setIconOn()
                    if camera.isCurrent():
                        self.mw.cameraPanel.setTabsEnabled(True)

                player.lock()

                if player.needs_render:
                    player.needs_render = False
                else:
                    continue

                rect = self.mw.pm.displayRect(player.uri, self.buffer.size())
                if player.image:
                    painter.drawImage(rect, player.image)

                x = rect.x()
                y = rect.y()
                w = rect.width()
                h = rect.height()

                if not (player.analyze_video or player.analyze_audio) and player.alarm_state:
                    player.setAlarmState(0)

                # display blinking alarm signal
                if self.mw.settingsPanel.alarm.chkShowDisplay.isChecked():
                    blinker_size = 20
                    rectBlinker = QRectF(x+w-1.25*blinker_size, y+h-1.25*blinker_size, blinker_size, blinker_size)
                    if camera := self.mw.cameraPanel.getCamera(player.uri):
                        if camera.isRecording():
                            if camera.isAlarming():
                                painter.drawImage(rectBlinker, self.alarm_recording.currentImage())
                            else:
                                if not player.systemTabSettings().record_always or not player.systemTabSettings().record_enable:
                                    painter.drawImage(rectBlinker, self.plain_recording.currentImage())
                        else:
                            if camera.isAlarming():
                                painter.drawImage(rectBlinker, QImage("image:alarm_plain.png"))

                # show model detection boxes
                if player.boxes is not None:
                    show = False
                    if self.mw.settingsPanel.proxy.generateAlarmsLocally() and camera.videoModelSettings:
                            show = camera.videoModelSettings.show
                    if show and len(player.boxes) and player.analyze_video:
                        scalex = w / player.image.rect().width()
                        scaley = h / player.image.rect().height()
                        painter.setPen(QColorConstants.Red)
                        for box in player.boxes:
                            p = (box[0] * scalex + x)
                            q = (box[1] * scaley + y)
                            r = (box[2] - box[0]) * scalex
                            s = (box[3] - box[1]) * scaley
                            painter.drawRect(QRectF(p, q, r, s).intersected(rect))

                if player.save_image_filename:
                    player.image.save(player.save_image_filename)
                    player.save_image_filename = None

                if self.isFocusedURI(player.uri) and player.isCameraStream():
                    painter.setPen(QColorConstants.White)
                    painter.drawRect(rect.adjusted(1, 1, -2, -2))

                if player.packet_drop_frame_counter > 0:
                    pen = QPen(QColorConstants.Yellow)
                    pen.setWidth(3)
                    painter.setPen(pen)
                    painter.drawRect(rect.adjusted(3, 3, -5, -5))

                player.unlock()

                blanks = self.mw.pm.getBlankSpace(self.buffer.size())
                for blank in blanks:
                    painter.fillRect(blank, QColorConstants.Black)

            self.mw.pm.unlock()

            for timer in self.mw.timers.values():

                if timer.attempting_reconnect:
                    if camera := self.mw.cameraPanel.getCamera(timer.uri):
                        camera.setIcon(QIcon(self.spinner.currentPixmap()))
                        if camera.isCurrent():
                            self.mw.cameraPanel.setTabsEnabled(False)
                        if camera.displayProfileIndex() != camera.recordProfileIndex():
                            mainState = camera.getStreamState(camera.recordProfileIndex())
                            subState = camera.getStreamState(camera.displayProfileIndex())
                            if subState == StreamState.CONNECTING and mainState == StreamState.CONNECTING:
                                if recordProfile := camera.getRecordProfile():
                                    if timer.uri == recordProfile.uri():
                                        continue
                    
                    timer.lock()
                    rect = self.mw.pm.displayRect(timer.uri, self.buffer.size())
                    painter.fillRect(rect, QColorConstants.Black)
                    painter.setPen(QColorConstants.LightGray)
                    rectSpinner = QRectF(0, 0, 40, 40)
                    rectSpinner.moveCenter(rect.center())
                    painter.drawImage(rectSpinner, self.spinner.currentImage())

                    day = timer.disconnected_time.strftime("%m/%d/%Y")
                    if day == datetime.now().strftime("%m/%d/%Y"):
                        day = "today"

                    text = f'{self.mw.getCameraName(timer.uri)}'
                    rectText = self.getTextRect(painter, text)
                    rectText.moveCenter(QPointF(rect.center().x(), rectSpinner.top()-3*rectText.height()))
                    painter.drawText(rectText, text)

                    text = f'Disconnected {day} at {timer.disconnected_time.strftime("%H:%M:%S")} ({self.interval2string(datetime.now() - timer.disconnected_time)})'
                    rectText = self.getTextRect(painter, text)
                    rectText.moveCenter(QPointF(rect.center().x(), rectSpinner.top()-2*rectText.height()))
                    painter.drawText(rectText, text)

                    text = f'Next reconnect attempt in {int(timer.remainingTime()/1000)} seconds'
                    rectText = self.getTextRect(painter, text)
                    rectText.moveCenter(QPointF(rectSpinner.center().x(), rectSpinner.top()-rectText.height()))
                    painter.drawText(rectText, text)

                    if self.isFocusedURI(timer.uri) and rect.isValid():
                        painter.setPen(QColorConstants.White)
                        painter.drawRect(rect.adjusted(1, 1, -2, -2))

                    timer.unlock()

        except Exception as ex:
            logger.error(f'GLWidget buildImage exception: {str(ex)}')
            logger.debug(traceback.format_exc())

    def getTextRect(self, painter, text):
        # to get exact bounding box, first estimate
        estimate = painter.fontMetrics().boundingRect(text)
        return painter.fontMetrics().boundingRect(estimate, 0, text).toRectF()

    def interval2string(self, interval):
        time_interval = int(interval.total_seconds())
        hours = int(time_interval / 3600)
        minutes = int ((time_interval - (hours * 3600)) / 60)
        seconds = int ((time_interval - (hours * 3600) - (minutes * 60)))
        buf = ""
        if hours > 0:
            buf += f'{hours} hours ' 
        if minutes > 0:
            buf += f'{minutes} minutes '
        buf += f'{seconds} seconds'
        return buf



#*******************************************************************************
# Cayenue/cayenue/components/infodialog.py
#
# Copyright (c) 2026 Stephen Rhodes 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/
from PyQt6.QtWidgets import QGridLayout, QLabel, \
    QDialog, QPushButton
from PyQt6.QtCore import Qt

class InfoDialog(QDialog):
    def __init__(self, p):
        super().__init__(p)
        self.setWindowTitle("info")
        self.lblMessage = QLabel()
        self.btnOK = QPushButton("OK")
        self.btnOK.clicked.connect(self.hide)
        lytMain = QGridLayout(self)
        lytMain.addWidget(self.lblMessage, 0, 0, 1, 1)
        lytMain.addWidget(QLabel(),        1, 0, 1, 1)
        lytMain.addWidget(self.btnOK,      2, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
